# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Dict, Iterator, Union
from urllib.parse import unquote, urlparse, urlunparse

import requests

from django.conf import settings
from django.http import QueryDict
from django.http.request import split_domain_port, validate_host
from django.http.response import StreamingHttpResponse
from rest_framework.decorators import renderer_classes
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request
from rest_framework.response import Response

from swh.model.hashutil import hash_to_hex
from swh.model.model import Sha1Git
from swh.model.swhids import ExtendedObjectType, ExtendedSWHID
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route
from swh.web.api.renderers import PlainTextRenderer
from swh.web.config import get_config
from swh.web.utils import archive, strtobool

API_GRAPH_PERM = "swh.web.api.graph"


def _resolve_origin_swhid(swhid: str, origin_urls: Dict[Sha1Git, str]) -> str:
    """
    Resolve origin url from its swhid sha1 representation.
    """
    parsed_swhid = ExtendedSWHID.from_string(swhid)
    if parsed_swhid.object_type == ExtendedObjectType.ORIGIN:
        if parsed_swhid.object_id in origin_urls:
            return origin_urls[parsed_swhid.object_id]
        else:
            origin_info = list(
                archive.lookup_origins_by_sha1s([hash_to_hex(parsed_swhid.object_id)])
            )[0]
            assert origin_info is not None
            origin_urls[parsed_swhid.object_id] = origin_info["url"]
            return origin_info["url"]
    else:
        return swhid


def _resolve_origin_swhids_in_graph_response(
    response: requests.Response,
) -> Iterator[bytes]:
    """
    Resolve origin urls from their swhid sha1 representations in graph service
    responses.
    """
    content_type = response.headers["Content-Type"]
    origin_urls: Dict[Sha1Git, str] = {}
    if content_type == "application/x-ndjson":
        for line in response.iter_lines():
            swhids = json.loads(line.decode("utf-8"))
            processed_line = []
            for swhid in swhids:
                processed_line.append(_resolve_origin_swhid(swhid, origin_urls))
            yield (json.dumps(processed_line) + "\n").encode()
    elif content_type == "text/plain":
        for line in response.iter_lines():
            if not line:
                continue
            processed_line = []
            swhids = line.decode("utf-8").split(" ")
            for swhid in swhids:
                processed_line.append(_resolve_origin_swhid(swhid, origin_urls))
            yield (" ".join(processed_line) + "\n").encode()
    else:
        for line in response.iter_lines():
            yield line + b"\n"


@api_route(r"/graph/", "api-1-graph-doc")
@api_doc("/graph/", category="Miscellaneous")
def api_graph(request: Request) -> None:
    """
    .. http:get:: /api/1/graph/(graph_query)/

        Provide fast access to the graph representation of the Software Heritage
        archive.

        That endpoint acts as a proxy for the `Software Heritage Graph service
        <https://docs.softwareheritage.org/devel/swh-graph/index.html>`_.

        It provides fast access to the `graph representation
        <https://docs.softwareheritage.org/devel/swh-model/data-model.html#data-structure>`_
        of the Software Heritage archive.

        For more details please refer to the `Graph RPC API documentation
        <https://docs.softwareheritage.org/devel/swh-graph/api.html>`_.

        .. warning::
            That endpoint is not publicly available and requires authentication and
            special user permission in order to be able to request it.

        :param string graph_query: query to forward to the Software Heritage Graph
            archive (see its `documentation
            <https://docs.softwareheritage.org/devel/swh-graph/api.html>`_)
        :query boolean resolve_origins: extra parameter defined by that proxy enabling
            to resolve origin urls from their sha1 representations

        :statuscode 200: no error
        :statuscode 400: an invalid graph query has been provided
        :statuscode 404: provided graph node cannot be found

        **Examples:**

        .. parsed-literal::

            :swh_web_api:`graph/leaves/swh:1:dir:432d1b21c1256f7408a07c577b6974bbdbcc1323/`
            :swh_web_api:`graph/neighbors/swh:1:rev:f39d7d78b70e0f39facb1e4fab77ad3df5c52a35/`
            :swh_web_api:`graph/visit/nodes/swh:1:snp:40f9f177b8ab0b7b3d70ee14bbc8b214e2b2dcfc?direction=backward&resolve_origins=true`
            :swh_web_api:`graph/visit/edges/swh:1:snp:40f9f177b8ab0b7b3d70ee14bbc8b214e2b2dcfc?direction=backward&resolve_origins=true`

    """
    return None


@api_route(r"/graph/(?P<graph_query>.+)/", "api-1-graph")
@renderer_classes([JSONRenderer, PlainTextRenderer])
def api_graph_proxy(
    request: Request, graph_query: str
) -> Union[Response, StreamingHttpResponse]:
    domain, port = split_domain_port(request.get_host())
    if not (domain and validate_host(domain, settings.UNAUTHENTICATED_HOSTS)):
        # request does not come from an identified allowed host/network (see
        # 'unauthenticated_api_hosts' config entry), check for proper auth and
        # permission
        if not bool(request.user and request.user.is_authenticated):
            return Response("Authentication credentials were not provided.", status=401)
        if not request.user.has_perm(API_GRAPH_PERM):
            return Response(
                "You do not have permission to perform this action.", status=403
            )

    graph_config = get_config()["graph"]
    graph_query = unquote(graph_query)
    graph_query_url = graph_config["server_url"]
    graph_query_url += graph_query

    parsed_url = urlparse(graph_query_url)
    query_dict = QueryDict(parsed_url.query, mutable=True)
    query_dict.update(request.GET)

    # clamp max_edges query parameter according to authentication
    if request.user.is_staff:
        max_edges = graph_config["max_edges"]["staff"]
    elif request.user.is_authenticated:
        max_edges = graph_config["max_edges"]["user"]
    else:
        max_edges = graph_config["max_edges"]["anonymous"]
    query_dict["max_edges"] = min(
        max_edges, int(query_dict.get("max_edges", max_edges + 1))
    )

    if query_dict:
        graph_query_url = urlunparse(
            parsed_url._replace(query=query_dict.urlencode(safe="/;:"))
        )

    response = requests.get(graph_query_url, stream=True)

    if response.status_code != 200:
        return Response(
            response.content,
            status=response.status_code,
            content_type=response.headers["Content-Type"],
        )

    # graph stats and counter endpoint responses are not streamed
    if response.headers.get("Transfer-Encoding") != "chunked":
        return Response(
            response.json(),
            status=response.status_code,
            content_type=response.headers["Content-Type"],
        )
    # other endpoint responses are streamed
    else:
        resolve_origins = strtobool(request.GET.get("resolve_origins", "false"))
        if response.status_code == 200 and resolve_origins:
            response_stream = _resolve_origin_swhids_in_graph_response(response)
        else:
            response_stream = map(lambda line: line + b"\n", response.iter_lines())
        return StreamingHttpResponse(
            response_stream,
            status=response.status_code,
            content_type=response.headers["Content-Type"],
        )



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from itertools import product

from prometheus_client import Gauge
from prometheus_client.registry import CollectorRegistry

from django.db.models import Count, DurationField, ExpressionWrapper, F, Sum

from swh.web.save_code_now.models import (
    SAVE_REQUEST_ACCEPTED,
    SAVE_REQUEST_PENDING,
    SAVE_REQUEST_REJECTED,
    SAVE_TASK_FAILED,
    SAVE_TASK_NOT_CREATED,
    SAVE_TASK_NOT_YET_SCHEDULED,
    SAVE_TASK_RUNNING,
    SAVE_TASK_SCHEDULED,
    SAVE_TASK_SUCCEEDED,
    SaveOriginRequest,
)
from swh.web.save_code_now.origin_save import get_savable_visit_types
from swh.web.save_origin_webhooks.generic_receiver import SUPPORTED_FORGE_TYPES

SWH_WEB_METRICS_REGISTRY = CollectorRegistry(auto_describe=True)

SUBMITTED_SAVE_REQUESTS_METRIC = "swh_web_submitted_save_requests"

_submitted_save_requests_gauge = Gauge(
    name=SUBMITTED_SAVE_REQUESTS_METRIC,
    documentation="Number of submitted origin save requests",
    labelnames=["status", "visit_type"],
    registry=SWH_WEB_METRICS_REGISTRY,
)

SUBMITTED_SAVE_REQUESTS_FROM_WEBHOOKS_METRIC = (
    "swh_web_submitted_save_requests_from_webhooks"
)

_submitted_save_requests_from_webhooks_gauge = Gauge(
    name=SUBMITTED_SAVE_REQUESTS_FROM_WEBHOOKS_METRIC,
    documentation="Number of submitted origin save requests through forge webhook receivers",
    labelnames=["status", "webhook_origin"],
    registry=SWH_WEB_METRICS_REGISTRY,
)


ACCEPTED_SAVE_REQUESTS_METRIC = "swh_web_accepted_save_requests"

_accepted_save_requests_gauge = Gauge(
    name=ACCEPTED_SAVE_REQUESTS_METRIC,
    documentation="Number of accepted origin save requests",
    labelnames=["load_task_status", "visit_type"],
    registry=SWH_WEB_METRICS_REGISTRY,
)


# Metric on the delay of save code now request per status and visit_type. This is the
# time difference between the save code now is requested and the time it got ingested.
ACCEPTED_SAVE_REQUESTS_DELAY_METRIC = "swh_web_save_requests_delay_seconds"

_accepted_save_requests_delay_gauge = Gauge(
    name=ACCEPTED_SAVE_REQUESTS_DELAY_METRIC,
    documentation="Save Requests Duration",
    labelnames=["load_task_status", "visit_type"],
    registry=SWH_WEB_METRICS_REGISTRY,
)


def compute_save_requests_metrics() -> None:
    """Compute Prometheus metrics related to origin save requests:

    - Number of submitted origin save requests
    - Number of accepted origin save requests
    - Save Code Now requests delay between request time and actual time of ingestion

    """

    request_statuses = (
        SAVE_REQUEST_ACCEPTED,
        SAVE_REQUEST_REJECTED,
        SAVE_REQUEST_PENDING,
    )

    load_task_statuses = (
        SAVE_TASK_NOT_CREATED,
        SAVE_TASK_NOT_YET_SCHEDULED,
        SAVE_TASK_SCHEDULED,
        SAVE_TASK_SUCCEEDED,
        SAVE_TASK_FAILED,
        SAVE_TASK_RUNNING,
    )

    # for metrics, we want access to all visit types
    visit_types = get_savable_visit_types(privileged_user=True)

    labels_set = product(request_statuses, visit_types)

    for labels in labels_set:
        _submitted_save_requests_gauge.labels(*labels).set(0)

    labels_set = product(request_statuses, SUPPORTED_FORGE_TYPES)

    for labels in labels_set:
        _submitted_save_requests_from_webhooks_gauge.labels(*labels).set(0)

    labels_set = product(load_task_statuses, visit_types)

    for labels in labels_set:
        _accepted_save_requests_gauge.labels(*labels).set(0)

    duration_load_task_statuses = (
        SAVE_TASK_FAILED,
        SAVE_TASK_SUCCEEDED,
    )

    for labels in product(duration_load_task_statuses, visit_types):
        _accepted_save_requests_delay_gauge.labels(*labels).set(0)

    counts = (
        SaveOriginRequest.objects.values(
            "status",
            "loading_task_status",
            "visit_type",
            "from_webhook",
            "webhook_origin",
        )
        .order_by()
        .annotate(Count("id"))
    )
    for count in counts:
        _submitted_save_requests_gauge.labels(
            status=count["status"],
            visit_type=count["visit_type"],
        ).inc(count["id__count"])
        if count["status"] == "accepted":
            _accepted_save_requests_gauge.labels(
                load_task_status=count["loading_task_status"],
                visit_type=count["visit_type"],
            ).inc(count["id__count"])
        if count["from_webhook"]:
            _submitted_save_requests_from_webhooks_gauge.labels(
                status=count["status"],
                webhook_origin=count["webhook_origin"],
            ).inc(count["id__count"])

    accepted_save_requests_delays = (
        SaveOriginRequest.objects.values("loading_task_status", "visit_type")
        .filter(visit_date__gt=F("request_date"))
        .filter(loading_task_status__in=[SAVE_TASK_SUCCEEDED, SAVE_TASK_FAILED])
        .order_by()
        .annotate(
            delay=Sum(
                ExpressionWrapper(
                    F("visit_date") - F("request_date"), output_field=DurationField()
                )
            )
        )
    )
    for accepted_save_requests_delay in accepted_save_requests_delays:
        _accepted_save_requests_delay_gauge.labels(
            load_task_status=accepted_save_requests_delay["loading_task_status"],
            visit_type=accepted_save_requests_delay["visit_type"],
        ).inc(accepted_save_requests_delay["delay"].total_seconds())



# Copyright (C) 2015-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import functools
from functools import wraps
import os
import re
import textwrap
from typing import List

import docutils.nodes
import docutils.parsers.rst
import docutils.utils

from django.shortcuts import redirect
from rest_framework.decorators import api_view

from swh.web.api.apiresponse import make_api_response
from swh.web.api.apiurls import CategoryId, api_urls
from swh.web.utils import parse_rst, reverse


class _HTTPDomainDocVisitor(docutils.nodes.NodeVisitor):
    """
    docutils visitor for walking on a parsed docutils document containing sphinx
    httpdomain roles. Its purpose is to extract relevant info regarding swh
    api endpoints (for instance url arguments) from their docstring written
    using sphinx httpdomain; and produce the main description back into a ReST
    string
    """

    # httpdomain roles we want to parse (based on sphinxcontrib.httpdomain 1.6)
    parameter_roles = ("param", "parameter", "arg", "argument")

    request_json_object_roles = ("reqjsonobj", "reqjson", "<jsonobj", "<json")

    request_json_array_roles = ("reqjsonarr", "<jsonarr")

    response_json_object_roles = ("resjsonobj", "resjson", ">jsonobj", ">json")

    response_json_array_roles = ("resjsonarr", ">jsonarr")

    query_parameter_roles = ("queryparameter", "queryparam", "qparam", "query")

    request_header_roles = ("<header", "reqheader", "requestheader")

    response_header_roles = (">header", "resheader", "responseheader")

    status_code_roles = ("statuscode", "status", "code")

    def __init__(self, document, data):
        super().__init__(document)
        self.data = data
        self.args_set = set()
        self.params_set = set()
        self.inputs_set = set()
        self.returns_set = set()
        self.status_codes_set = set()
        self.reqheaders_set = set()
        self.resheaders_set = set()
        self.current_json_obj = None
        self.current_field_name = None

    def _default_visit(self, node: docutils.nodes.Element) -> str:
        """Simply visits a text node, drops its start and end tags, visits
        the children, and concatenates their results."""
        return "".join(map(self.dispatch_visit, node.children))

    def visit_emphasis(self, node: docutils.nodes.emphasis) -> str:
        return f"*{self._default_visit(node)}*"

    def visit_strong(self, node: docutils.nodes.strong) -> str:
        return f"**{self._default_visit(node)}**"

    def visit_reference(self, node: docutils.nodes.reference) -> str:
        text = self._default_visit(node)
        refuri = node.attributes.get("refuri")
        if refuri is not None:
            return f"`{text} <{refuri}>`_"
        else:
            return text

    def visit_target(self, node: docutils.nodes.target) -> str:
        return ""

    def visit_literal(self, node: docutils.nodes.literal) -> str:
        return f"``{self._default_visit(node)}``"

    def visit_field_name(self, node: docutils.nodes.field_name) -> str:
        self.current_field_name = node.astext()
        return ""

    def visit_field_body(self, node: docutils.nodes.field_body) -> str:
        text = self._default_visit(node).strip()
        assert text, str(node)
        field_data = self.current_field_name.split(" ")
        # Parameters
        if field_data[0] in self.parameter_roles:
            if field_data[2] not in self.args_set:
                self.data["args"].append(
                    {"name": field_data[2], "type": field_data[1], "doc": text}
                )
                self.args_set.add(field_data[2])
        # Query Parameters
        if field_data[0] in self.query_parameter_roles:
            if field_data[2] not in self.params_set:
                self.data["params"].append(
                    {"name": field_data[2], "type": field_data[1], "doc": text}
                )
                self.params_set.add(field_data[2])
        # Request data type
        if (
            field_data[0] in self.request_json_array_roles
            or field_data[0] in self.request_json_object_roles
        ):
            # array
            if field_data[0] in self.request_json_array_roles:
                self.data["input_type"] = "array"
            # object
            else:
                self.data["input_type"] = "object"
            # input object field
            if field_data[2] not in self.inputs_set:
                self.data["inputs"].append(
                    {"name": field_data[2], "type": field_data[1], "doc": text}
                )
                self.inputs_set.add(field_data[2])
                self.current_json_obj = self.data["inputs"][-1]
        # Response type
        if (
            field_data[0] in self.response_json_array_roles
            or field_data[0] in self.response_json_object_roles
        ):
            # array
            if field_data[0] in self.response_json_array_roles:
                self.data["return_type"] = "array"
            # object
            else:
                self.data["return_type"] = "object"
            # returned object field
            if field_data[2] not in self.returns_set:
                self.data["returns"].append(
                    {"name": field_data[2], "type": field_data[1], "doc": text}
                )
                self.returns_set.add(field_data[2])
                self.current_json_obj = self.data["returns"][-1]
        # Status Codes
        if field_data[0] in self.status_code_roles:
            if field_data[1] not in self.status_codes_set:
                self.data["status_codes"].append({"code": field_data[1], "doc": text})
                self.status_codes_set.add(field_data[1])
        # Request Headers
        if field_data[0] in self.request_header_roles:
            if field_data[1] not in self.reqheaders_set:
                self.data["reqheaders"].append({"name": field_data[1], "doc": text})
                self.reqheaders_set.add(field_data[1])
        # Response Headers
        if field_data[0] in self.response_header_roles:
            if field_data[1] not in self.resheaders_set:
                resheader = {"name": field_data[1], "doc": text}
                self.data["resheaders"].append(resheader)
                self.resheaders_set.add(field_data[1])
                if (
                    resheader["name"] == "Content-Type"
                    and resheader["doc"] == "application/octet-stream"
                ):
                    self.data["return_type"] = "octet stream"

        # Don't return anything in the description; these nodes only add text
        # to other fields
        return ""

    # We ignore these nodes and handle their subtrees directly in
    # visit_field_name and visit_field_body
    visit_field = visit_field_list = _default_visit

    def visit_paragraph(self, node: docutils.nodes.paragraph) -> str:
        """
        Visit relevant paragraphs to parse
        """
        # only parsed top level paragraphs
        text = self._default_visit(node)

        return "\n\n" + text

    def visit_literal_block(self, node: docutils.nodes.literal_block) -> str:
        """
        Visit literal blocks
        """

        text = node.astext()
        node_classes = node["classes"]
        if node_classes and node_classes[0] == "code":
            language = node_classes[1] if len(node_classes) > 1 else ""
            # code block will be rendered by highlightjs
            return (
                f'\n\n.. raw:: html\n\n    <pre><code class="{language}">'
                f'{textwrap.indent(text, " " * 4)[4:]}</code></pre>'
            )
        else:
            return f"\n\n::\n\n{textwrap.indent(text, '   ')}\n"

    def visit_bullet_list(self, node: docutils.nodes.bullet_list) -> str:
        parts = ["\n\n"]
        for child in node.findall():
            # process list item
            if isinstance(child, docutils.nodes.list_item):
                line_text = self.dispatch_visit(child)
                parts.append("\t* %s\n" % textwrap.indent(line_text, "\t  ").strip())

        return "".join(parts)

    # visit_bullet_list collects and handles this with a more global view:
    visit_list_item = _default_visit

    def visit_warning(self, node: docutils.nodes.warning) -> str:
        text = self._default_visit(node)
        return "\n\n.. warning::\n%s\n" % textwrap.indent(text, "\t")

    def visit_note(self, node: docutils.nodes.note) -> str:
        text = self._default_visit(node)
        return "\n\n.. note::\n%s\n" % textwrap.indent(text, "\t")

    def visit_important(self, node: docutils.nodes.important) -> str:
        text = self._default_visit(node)
        return "\n\n.. important::\n%s\n" % textwrap.indent(text, "\t")

    def visit_Text(self, node: docutils.nodes.Text) -> str:
        """Leaf node"""
        return str(node).replace("\n", " ")  # Prettier in generated HTML

    def visit_problematic(self, node: docutils.nodes.problematic) -> str:
        # api urls cleanup to generate valid links afterwards
        text = self._default_visit(node)
        subs_made = 1
        while subs_made:
            (text, subs_made) = re.subn(r"(:http:.*)(\(\w+\))", r"\1", text)
        subs_made = 1
        while subs_made:
            (text, subs_made) = re.subn(r"(:http:.*)(\[.*\])", r"\1", text)
        text = re.sub(r"([^:])//", r"\1/", text)
        # transform references to api endpoints doc into valid rst links
        text = re.sub(":http:get:`([^,`]*)`", r"`\1 <\1doc/>`_", text)
        text = re.sub(":http:post:`([^,`]*)`", r"`\1 <\1doc/>`_", text)
        # transform references to some elements into bold text
        text = re.sub(":http:header:`(.*)`", r"**\1**", text)
        text = re.sub(":func:`(.*)`", r"**\1**", text)
        text = re.sub(":mod:`(.*)`", r"**\1**", text)

        # extract example urls
        if ":swh_web_api:" in text:
            # Extract examples to their own section
            examples_str = re.sub(":swh_web_api:`(.+)`.*", r"/api/1/\1", text)
            self.data["examples"] += examples_str.split("\n")
        return text

    def visit_block_quote(self, node: docutils.nodes.block_quote) -> str:
        return self._default_visit(node)

    def visit_inline(self, node):
        return self._default_visit(node)

    def visit_code_block(self, node) -> str:
        return self._default_visit(node)

    def visit_title_reference(self, node: docutils.nodes.title_reference) -> str:
        text = self._default_visit(node)
        raise Exception(
            f"Unexpected title reference. "
            f"Possible cause: you used `{text}` instead of ``{text}``"
        )

    def visit_document(self, node: docutils.nodes.document) -> None:
        text = self._default_visit(node)

        # Strip examples; they are displayed separately
        text = re.split("\n\\*\\*Examples?:\\*\\*\n", text)[0]

        self.data["description"] = text.strip()

    def visit_system_message(self, node):
        return ""

    def unknown_visit(self, node) -> str:
        raise NotImplementedError(
            f"Unknown node type: {node.__class__.__name__}. Value: {node}"
        )

    def unknown_departure(self, node):
        pass


def _parse_httpdomain_doc(doc, data):
    doc_lines = doc.split("\n")
    doc_lines_filtered = []
    urls = defaultdict(list)
    default_http_methods = ["HEAD", "OPTIONS"]
    # httpdomain is a sphinx extension that is unknown to docutils but
    # fortunately we can still parse its directives' content,
    # so remove lines with httpdomain directives before executing the
    # rst parser from docutils
    for doc_line in doc_lines:
        if ".. http" not in doc_line:
            doc_lines_filtered.append(doc_line)
        else:
            url = doc_line[doc_line.find("/") :]
            # emphasize url arguments for html rendering
            url = re.sub(r"\((\w+)\)", r" **\(\1\)** ", url)
            method = re.search(r"http:(\w+)::", doc_line).group(1)
            urls[url].append(method.upper())

    for url, methods in urls.items():
        data["urls"].append({"rule": url, "methods": methods + default_http_methods})
    # parse the rst docstring and do not print system messages about
    # unknown httpdomain roles
    document = parse_rst("\n".join(doc_lines_filtered), report_level=5)
    # remove the system_message nodes from the parsed document
    for node in document.findall(docutils.nodes.system_message):
        node.parent.remove(node)
    # visit the document nodes to extract relevant endpoint info
    visitor = _HTTPDomainDocVisitor(document, data)
    document.walkabout(visitor)


class APIDocException(Exception):
    """
    Custom exception to signal errors in the use of the APIDoc decorators
    """


def api_doc(
    route: str,
    *,
    category: CategoryId,
    noargs: bool = False,
    tags: List[str] = [],
    api_version: str = "1",
):
    """
    Decorator for an API endpoint implementation used to generate a dedicated
    view displaying its HTML documentation.

    The documentation will be generated from the endpoint docstring based on
    sphinxcontrib-httpdomain format.

    Args:
        route: documentation page's route
        noargs: set to True if the route has no arguments, and its
            result should be displayed anytime its documentation
            is requested. Default to False
        tags: Further information on api endpoints. Two values are
            possibly expected:

                * hidden: remove the entry points from the listing
                * upcoming: display the entry point but it is not followable
                * deprecated: display the entry point as deprecated in the index
        api_version: api version string
    """

    tags_set = set(tags)

    # @api_doc() Decorator call
    def decorator(f):
        # if the route is not hidden, add it to the index
        if "hidden" not in tags_set:
            doc_data = get_doc_data(f, route, noargs)
            doc_desc = doc_data["description"]
            api_urls.add_doc_route(
                route,
                category,
                re.split(r"\.\s", doc_desc)[0],
                noargs=noargs,
                api_version=api_version,
                tags=tags_set,
            )

        # create a dedicated view to display endpoint HTML doc
        @api_view(["GET", "HEAD"])
        @wraps(f)
        def doc_view(request):
            doc_data = get_doc_data(f, route, noargs)
            return make_api_response(request, None, doc_data)

        route_name = "%s-doc" % route[1:-1].replace("/", "-")
        urlpattern = f"api/{api_version}{route}doc/"

        view_name = "api-%s-%s" % (api_version, route_name)
        api_urls.add_url_pattern(urlpattern, doc_view, view_name)

        # for backward compatibility as previous apidoc URLs were missing
        # the /api prefix
        old_view_name = view_name.replace("api-", "")
        old_urlpattern = f"{api_version}{route}doc/"

        @api_view(["GET", "HEAD"])
        def old_doc_view(request):
            return redirect(reverse(view_name))

        api_urls.add_url_pattern(old_urlpattern, old_doc_view, old_view_name)

        @wraps(f)
        def documented_view(request, **kwargs):
            doc_data = get_doc_data(f, route, noargs)
            try:
                return {"data": f(request, **kwargs), "doc_data": doc_data}
            except Exception as exc:
                exc.doc_data = doc_data
                raise exc

        return documented_view

    return decorator


@functools.lru_cache(maxsize=32)
def get_doc_data(f, route, noargs):
    """
    Build documentation data for the decorated api endpoint function
    """
    data = {
        "description": "",
        "response_data": None,
        "urls": [],
        "args": [],
        "params": [],
        "input_type": "",
        "inputs": [],
        "resheaders": [],
        "reqheaders": [],
        "return_type": "",
        "returns": [],
        "status_codes": [],
        "examples": [],
        "route": route,
        "noargs": noargs,
    }

    if not f.__doc__:
        raise APIDocException(
            "apidoc: expected a docstring" " for function %s" % (f.__name__,)
        )

    # use raw docstring as endpoint documentation if sphinx
    # httpdomain is not used
    if ".. http" not in f.__doc__:
        data["description"] = f.__doc__
    # else parse the sphinx httpdomain docstring with docutils
    # (except when building the swh-web documentation through autodoc
    # sphinx extension, not needed and raise errors with sphinx >= 1.7)
    elif "SWH_DOC_BUILD" not in os.environ:
        _parse_httpdomain_doc(f.__doc__, data)
        # process input/returned object info for nicer html display
        inputs_list = ""
        returns_list = ""
        for inp in data["inputs"]:
            # special case for array of non object type, for instance
            # :<jsonarr string -: an array of string
            if inp["name"] != "-":
                inputs_list += "\t* **%s (%s)**: %s\n" % (
                    inp["name"],
                    inp["type"],
                    textwrap.indent(inp["doc"], "\t  "),
                )
        for ret in data["returns"]:
            # special case for array of non object type, for instance
            # :>jsonarr string -: an array of string
            if ret["name"] != "-":
                returns_list += "\t* **%s (%s)**: %s\n" % (
                    ret["name"],
                    ret["type"],
                    textwrap.indent(ret["doc"], "\t  "),
                )
        data["inputs_list"] = inputs_list
        data["returns_list"] = returns_list

    return data


DOC_COMMON_HEADERS = """
        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept`
            header of request"""
DOC_RESHEADER_LINK = """
        :resheader Link: indicates that a subsequent result page is
            available and contains the url pointing to it
"""

DEFAULT_SUBSTITUTIONS = {
    "common_headers": DOC_COMMON_HEADERS,
    "resheader_link": DOC_RESHEADER_LINK,
}


def format_docstring(**substitutions):
    def decorator(f):
        f.__doc__ = f.__doc__.format(**{**DEFAULT_SUBSTITUTIONS, **substitutions})
        return f

    return decorator



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.web.api.apiurls import api_route
from swh.web.tests.helpers import check_api_get_responses
from swh.web.utils import reverse


@api_route(r"/some/route/(?P<int_arg>[0-9]+)/", "api-1-some-route")
def api_some_route(request, int_arg):
    return {"result": int(int_arg)}


@api_route(
    r"/never/cache/route/(?P<int_arg>[0-9]+)/",
    "api-1-never-cache-route",
    never_cache=True,
)
def api_never_cache_route(request, int_arg):
    return {"result": int(int_arg)}


@api_route(
    r"/never/cache/route/error/",
    "api-1-never-cache-route-with-error",
    never_cache=True,
)
def api_never_cache_route_with_error(request):
    raise Exception("error")


def test_api_route_with_cache(api_client):
    url = reverse("api-1-some-route", url_args={"int_arg": 1})
    resp = check_api_get_responses(api_client, url, status_code=200)
    assert resp.data == {"result": 1}
    assert "Cache-Control" not in resp


_cache_control = "max-age=0, no-cache, no-store, must-revalidate"


def test_api_route_never_cache(api_client):
    url = reverse("api-1-never-cache-route", url_args={"int_arg": 1})
    resp = check_api_get_responses(api_client, url, status_code=200)
    assert resp.data == {"result": 1}
    assert "Cache-Control" in resp
    assert resp["Cache-Control"].startswith(_cache_control)


def test_api_route_never_cache_with_error(api_client):
    url = reverse("api-1-never-cache-route-with-error")
    resp = check_api_get_responses(api_client, url, status_code=500)
    assert "Cache-Control" in resp
    assert resp["Cache-Control"].startswith(_cache_control)



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import textwrap

import docutils
import pytest

from rest_framework.response import Response

from swh.storage.exc import StorageAPIError, StorageDBError
from swh.web.api.apidoc import _parse_httpdomain_doc, api_doc
from swh.web.api.apiurls import api_route
from swh.web.tests.django_asserts import assert_contains, assert_not_contains
from swh.web.tests.helpers import (
    check_api_get_responses,
    check_html_get_response,
    prettify_html,
)
from swh.web.utils import reverse
from swh.web.utils.exc import BadInputExc, ForbiddenExc, NotFoundExc

_httpdomain_doc = """
.. http:get:: /api/1/revision/(sha1_git)/

    Get information about a revision in the archive.
    Revisions are identified by **sha1** checksums, compatible with Git commit
    identifiers.
    See :func:`swh.model.git_objects.revision_git_object` in our data model
    module for details about how they are computed.

    :param string sha1_git: hexadecimal representation of the revision
        **sha1_git** identifier

    :reqheader Accept: the requested response content type,
        either ``application/json`` (default) or ``application/yaml``
    :resheader Content-Type: this depends on :http:header:`Accept` header
        of request

    :<json int n: sample input integer
    :<json string s: sample input string
    :<json array a: sample input array

    :>json object author: information about the author of the revision
    :>json object committer: information about the committer of the revision
    :>json string committer_date: RFC3339 representation of the commit date
    :>json string date: RFC3339 representation of the revision date
    :>json string directory: the unique identifier that revision points to
    :>json string directory_url: link to
        :http:get:`/api/1/directory/(sha1_git)/[(path)/]` to get information
        about the directory associated to the revision
    :>json string id: the revision unique identifier
    :>json boolean merge: whether or not the revision corresponds to a merge
        commit
    :>json string message: the message associated to the revision
    :>json array parents: the parents of the revision, i.e. the previous
        revisions that head directly to it, each entry of that array contains
        an unique parent revision identifier but also a link to
        :http:get:`/api/1/revision/(sha1_git)/` to get more information
        about it
    :>json string type: the type of the revision

    :statuscode 200: no error
    :statuscode 400: an invalid **sha1_git** value has been provided
    :statuscode 404: requested revision cannot be found in the archive

    **Example:**

    .. parsed-literal::

        :swh_web_api:`revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/`
"""


_exception_http_code = {
    BadInputExc: 400,
    ForbiddenExc: 403,
    NotFoundExc: 404,
    Exception: 500,
    StorageAPIError: 503,
    StorageDBError: 503,
}


def test_apidoc_nodoc_failure():
    with pytest.raises(Exception):

        @api_doc("/my/nodoc/url/", "test")
        def apidoc_nodoc_tester(request, arga=0, argb=0):
            return Response(arga + argb)


@api_route(r"/some/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/", "api-1-some-doc-route")
@api_doc("/some/doc/route/", category="test")
def apidoc_route(request, myarg, myotherarg, akw=0):
    """
    Sample doc
    """
    return {"result": int(myarg) + int(myotherarg) + akw}


def test_apidoc_route_doc(client):
    api_view_name = "api-1-some-doc-route-doc"
    doc_url = reverse(api_view_name)
    assert doc_url == "/api/1/some/doc/route/doc/"
    resp = check_html_get_response(
        client, doc_url, status_code=200, template_used="apidoc.html"
    )

    # check apidoc breadcrumbs links
    api_view_name_split = api_view_name.split("-")
    for i in range(2, len(api_view_name_split) - 1):
        sub_doc_url = "/" + ("/".join(api_view_name_split[:i])) + "/doc/"
        assert_not_contains(resp, f'<a href="{sub_doc_url}">')
    assert_contains(resp, f'<a href="{doc_url}">')

    # check previous erroneous URL now redirects to the fixed one
    url = reverse("1-some-doc-route-doc")
    resp = check_html_get_response(client, url, status_code=302)
    assert resp["location"] == doc_url


def test_apidoc_route_fn(api_client):
    url = reverse("api-1-some-doc-route", url_args={"myarg": 1, "myotherarg": 1})
    check_api_get_responses(api_client, url, status_code=200)


@api_route(r"/test/error/(?P<exc_name>.+)/", "api-1-test-error")
@api_doc("/test/error/", category="test")
def apidoc_test_error_route(request, exc_name):
    """
    Sample doc
    """
    for e in _exception_http_code.keys():
        if e.__name__ == exc_name:
            raise e("Error")


def test_apidoc_error(api_client):
    for exc, code in _exception_http_code.items():
        url = reverse("api-1-test-error", url_args={"exc_name": exc.__name__})
        check_api_get_responses(api_client, url, status_code=code)


@api_route(
    r"/some/full/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/",
    "api-1-some-complete-doc-route",
)
@api_doc("/some/complete/doc/route/", category="test")
def apidoc_full_stack(request, myarg, myotherarg, akw=0):
    """
    Sample doc
    """
    return {"result": int(myarg) + int(myotherarg) + akw}


def test_apidoc_full_stack_doc(client):
    url = reverse("api-1-some-complete-doc-route-doc")
    check_html_get_response(client, url, status_code=200, template_used="apidoc.html")


def test_apidoc_full_stack_fn(api_client):
    url = reverse(
        "api-1-some-complete-doc-route", url_args={"myarg": 1, "myotherarg": 1}
    )
    check_api_get_responses(api_client, url, status_code=200)


@api_route(r"/test/post/only/", "api-1-test-post-only", methods=["POST"])
@api_doc("/test/post/only/", category="test")
def apidoc_test_post_only(request, exc_name):
    """
    Sample doc
    """
    return {"result": "some data"}


def test_apidoc_post_only(client):
    # a dedicated view accepting GET requests should have
    # been created to display the HTML documentation
    url = reverse("api-1-test-post-only-doc")
    check_html_get_response(client, url, status_code=200, template_used="apidoc.html")


def test_api_doc_parse_httpdomain():
    doc_data = {
        "description": "",
        "urls": [],
        "args": [],
        "params": [],
        "resheaders": [],
        "reqheaders": [],
        "input_type": "",
        "inputs": [],
        "return_type": "",
        "returns": [],
        "status_codes": [],
        "examples": [],
    }

    _parse_httpdomain_doc(_httpdomain_doc, doc_data)

    expected_urls = [
        {
            "rule": "/api/1/revision/ **\\(sha1_git\\)** /",
            "methods": ["GET", "HEAD", "OPTIONS"],
        }
    ]

    assert "urls" in doc_data
    assert doc_data["urls"] == expected_urls

    expected_description = (
        "Get information about a revision in the archive. "
        "Revisions are identified by **sha1** checksums, "
        "compatible with Git commit identifiers. See "
        "**swh.model.git_objects.revision_git_object** in "
        "our data model module for details about how they "
        "are computed."
    )

    assert "description" in doc_data
    assert doc_data["description"] == expected_description

    expected_args = [
        {
            "name": "sha1_git",
            "type": "string",
            "doc": (
                "hexadecimal representation of the revision " "**sha1_git** identifier"
            ),
        }
    ]

    assert "args" in doc_data
    assert doc_data["args"] == expected_args

    expected_params = []
    assert "params" in doc_data
    assert doc_data["params"] == expected_params

    expected_reqheaders = [
        {
            "doc": (
                "the requested response content type, either "
                "``application/json`` (default) or ``application/yaml``"
            ),
            "name": "Accept",
        }
    ]

    assert "reqheaders" in doc_data
    assert doc_data["reqheaders"] == expected_reqheaders

    expected_resheaders = [
        {"doc": "this depends on **Accept** header of request", "name": "Content-Type"}
    ]

    assert "resheaders" in doc_data
    assert doc_data["resheaders"] == expected_resheaders

    expected_statuscodes = [
        {"code": "200", "doc": "no error"},
        {"code": "400", "doc": "an invalid **sha1_git** value has been provided"},
        {"code": "404", "doc": "requested revision cannot be found in the archive"},
    ]

    assert "status_codes" in doc_data
    assert doc_data["status_codes"] == expected_statuscodes

    expected_input_type = "object"

    assert "input_type" in doc_data
    assert doc_data["input_type"] == expected_input_type

    expected_inputs = [
        {"name": "n", "type": "int", "doc": "sample input integer"},
        {"name": "s", "type": "string", "doc": "sample input string"},
        {"name": "a", "type": "array", "doc": "sample input array"},
    ]

    assert "inputs" in doc_data
    assert doc_data["inputs"] == expected_inputs

    expected_return_type = "object"

    assert "return_type" in doc_data
    assert doc_data["return_type"] == expected_return_type

    expected_returns = [
        {
            "name": "author",
            "type": "object",
            "doc": "information about the author of the revision",
        },
        {
            "name": "committer",
            "type": "object",
            "doc": "information about the committer of the revision",
        },
        {
            "name": "committer_date",
            "type": "string",
            "doc": "RFC3339 representation of the commit date",
        },
        {
            "name": "date",
            "type": "string",
            "doc": "RFC3339 representation of the revision date",
        },
        {
            "name": "directory",
            "type": "string",
            "doc": "the unique identifier that revision points to",
        },
        {
            "name": "directory_url",
            "type": "string",
            "doc": (
                "link to `/api/1/directory/ </api/1/directory/doc/>`_ "
                "to get information about the directory associated to "
                "the revision"
            ),
        },
        {"name": "id", "type": "string", "doc": "the revision unique identifier"},
        {
            "name": "merge",
            "type": "boolean",
            "doc": "whether or not the revision corresponds to a merge commit",
        },
        {
            "name": "message",
            "type": "string",
            "doc": "the message associated to the revision",
        },
        {
            "name": "parents",
            "type": "array",
            "doc": (
                "the parents of the revision, i.e. the previous revisions "
                "that head directly to it, each entry of that array "
                "contains an unique parent revision identifier but also a "
                "link to `/api/1/revision/ </api/1/revision/doc/>`_ "
                "to get more information about it"
            ),
        },
        {"name": "type", "type": "string", "doc": "the type of the revision"},
    ]

    assert "returns" in doc_data
    assert doc_data["returns"] == expected_returns

    expected_examples = ["/api/1/revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/"]

    assert "examples" in doc_data
    assert doc_data["examples"] == expected_examples


@api_route(r"/post/endpoint/", "api-1-post-endpoint", methods=["POST"])
@api_doc("/post/endpoint/", category="test")
def apidoc_test_post_endpoint(request):
    """
    .. http:post:: /api/1/post/endpoint/

        Endpoint documentation

        :<jsonarr string -: Input array of SWHIDs

        :>json object <swhid>: an object whose keys are input SWHIDs
            and values objects with the following keys:

                * **known (bool)**: whether the object was found

    """
    pass


def test_apidoc_input_output_doc(client):
    url = reverse("api-1-post-endpoint-doc")
    rv = check_html_get_response(
        client, url, status_code=200, template_used="apidoc.html"
    )

    input_html_doc = textwrap.indent(
        (
            '<dl class="row">\n'
            ' <dt class="col col-md-2 text-end">\n'
            "  array\n"
            " </dt>\n"
            ' <dd class="col col-md-9">\n'
            "  <p>\n"
            "   Input array of SWHIDs\n"
            "  </p>\n"
            " </dd>\n"
            "</dl>\n"
        ),
        " " * 8,
    )

    if docutils.__version_info__ >= (0, 17):
        output_html_doc = textwrap.indent(
            (
                '<dl class="row">\n'
                ' <dt class="col col-md-2 text-end">\n'
                "  object\n"
                " </dt>\n"
                ' <dd class="col col-md-9">\n'
                "  <p>\n"
                "   an object containing the following keys:\n"
                "  </p>\n"
                '  <div class="swh-rst">\n'
                "   <main>\n"
                "    <blockquote>\n"
                "     <ul>\n"
                "      <li>\n"
                "       <p>\n"
                "        <strong>\n"
                "         &lt;swhid&gt; (object)\n"
                "        </strong>\n"
                "        :           an object whose keys are input SWHIDs"
                " and values objects with the following keys:\n"
                "       </p>\n"
                "       <blockquote>\n"
                '        <ul class="simple">\n'
                "         <li>\n"
                "          <p>\n"
                "           <strong>\n"
                "            known (bool)\n"
                "           </strong>\n"
                "           : whether the object was found\n"
                "          </p>\n"
                "         </li>\n"
                "        </ul>\n"
                "       </blockquote>\n"
                "      </li>\n"
                "     </ul>\n"
                "    </blockquote>\n"
                "   </main>\n"
                "  </div>\n"
                " </dd>\n"
                "</dl>\n"
            ),
            " " * 8,
        )
    else:
        output_html_doc = textwrap.indent(
            (
                '<dl class="row">\n'
                ' <dt class="col col-md-2 text-end">\n'
                "  object\n"
                " </dt>\n"
                ' <dd class="col col-md-9">\n'
                "  <p>\n"
                "   an object containing the following keys:\n"
                "  </p>\n"
                '  <div class="swh-rst">\n'
                '   <div class="document">\n'
                "    <blockquote>\n"
                "     <ul>\n"
                "      <li>\n"
                "       <p>\n"
                "        <strong>\n"
                "         &lt;swhid&gt; (object)\n"
                "        </strong>\n"
                "        :           an object whose keys are input SWHIDs"
                " and values objects with the following keys:\n"
                "       </p>\n"
                "       <blockquote>\n"
                '        <ul class="simple">\n'
                "         <li>\n"
                "          <p>\n"
                "           <strong>\n"
                "            known (bool)\n"
                "           </strong>\n"
                "           : whether the object was found\n"
                "          </p>\n"
                "         </li>\n"
                "        </ul>\n"
                "       </blockquote>\n"
                "      </li>\n"
                "     </ul>\n"
                "    </blockquote>\n"
                "   </div>\n"
                "  </div>\n"
                " </dd>\n"
                "</dl>"
            ),
            " " * 8,
        )

    html = prettify_html(rv.content)

    assert input_html_doc in html
    assert output_html_doc in html


@api_route(r"/endpoint/links/in/doc/", "api-1-endpoint-links-in-doc")
@api_doc("/endpoint/links/in/doc/", category="test")
def apidoc_test_endpoint_with_links_in_doc(request):
    """
    .. http:get:: /api/1/post/endpoint/

        Endpoint documentation with links to
        :http:get:`/api/1/content/[(hash_type):](hash)/`,
        :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
        and `archive <https://archive.softwareheritage.org>`_.
    """
    pass


def test_apidoc_with_links(client):
    url = reverse("api-1-endpoint-links-in-doc")
    rv = check_html_get_response(
        client, url, status_code=200, template_used="apidoc.html"
    )

    html = prettify_html(rv.content)

    endpoint_full_url = textwrap.indent(
        "<p>\n http://testserver/api/1/post/endpoint/\n</p>", " " * 14
    )

    first_link = textwrap.indent(
        (
            '<a class="reference external" href="/api/1/content/doc/">\n'
            " /api/1/content/\n"
            "</a>"
        ),
        " " * 11,
    )

    second_link = textwrap.indent(
        (
            '<a class="reference external" href="/api/1/directory/doc/">\n'
            " /api/1/directory/\n"
            "</a>"
        ),
        " " * 11,
    )

    third_link = textwrap.indent(
        (
            '<a class="reference external" '
            'href="https://archive.softwareheritage.org">\n'
            " archive\n"
            "</a>"
        ),
        " " * 11,
    )

    assert endpoint_full_url in html
    assert first_link in html
    assert second_link in html
    assert third_link in html


@api_route(r"/endpoint/links-in-text/in/doc/", "api-1-endpoint-links-in-text-in-doc")
@api_doc("/endpoint/links/in/doc/", category="test")
def apidoc_test_endpoint_with_links_in_text_in_doc(request):
    """
    .. http:get:: /api/1/links-in-text/endpoint/

        Endpoint documentation with links to
        `archive <https://archive.softwareheritage.org>`_
        and `documentation <https://docs.softwareheritage.org>`_
        for testing.
    """
    pass


def test_apidoc_with_links_in_text(client):
    url = reverse("api-1-endpoint-links-in-text-in-doc")
    rv = check_html_get_response(
        client, url, status_code=200, template_used="apidoc.html"
    )

    html = prettify_html(rv.content)

    assert (
        textwrap.indent(
            (
                "<p>\n"
                " Endpoint documentation with links to\n"
                ' <a class="reference external" href="https://archive.softwareheritage.org">\n'
                "  archive\n"
                " </a>\n"
                " and\n"
                ' <a class="reference external" href="https://docs.softwareheritage.org">\n'
                "  documentation\n"
                " </a>\n"
                " for testing.\n"
                "</p>\n"
            ),
            " " * 10,
        )
        in html
    )


@api_route(r"/endpoint/codeblock-in-doc/", "api-1-endpoint-codeblock-in-doc")
@api_doc("/endpoint/links/in/doc/", category="test")
def apidoc_test_endpoint_with_codeblock_in_doc(request):
    """
    .. http:get:: /api/1/codeblock-in-doc/endpoint/

        .. code-block:: python

            def add(a, b):
                return a + B

    """
    pass


def test_apidoc_with_codeblock(client):
    url = reverse("api-1-endpoint-codeblock-in-doc")
    rv = check_html_get_response(
        client, url, status_code=200, template_used="apidoc.html"
    )

    html = prettify_html(rv.content)

    assert (
        '         <pre><code class="python">def add(a, b):\n'
        "    return a + B</code></pre>"
    ) in html



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from django import forms
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.paginator import Paginator
from django.forms.models import model_to_dict
from django.utils.translation import gettext as _

from .emails import send_alteration_magic_link, send_email_magic_link
from .models import (
    Alteration,
    AlterationCategory,
    BlockList,
    Event,
    EventCategory,
    EventRecipient,
    Origin,
    Token,
)
from .utils import generate_alteration_changelog, generate_origin_changelog

if TYPE_CHECKING:
    from django.core.paginator import Page
    from django.http import HttpRequest


class MultipleOriginField(forms.MultipleChoiceField):
    def validate(self, value):
        """Validate a list of origin."""
        from swh.web.utils.archive import lookup_origin

        if self.required and not value:
            raise ValidationError(self.error_messages["required"], code="required")
        # Validate that each value in the value list is a swh origin
        for origin in value:
            try:
                lookup_origin(origin)
            except ObjectDoesNotExist:  # Should be a utils.exc.NotFoundExc
                raise ValidationError(
                    _("%(origin)s is not archived by Software Heritage")
                    % {"origin": origin}
                )


class EmailVerificationForm(forms.Form):
    """Email verification form."""

    email = forms.EmailField(label=_("Email"), required=True)

    def __init__(self, *args, request: HttpRequest, **kwargs):
        """Store the extra ``request`` parameters.

        Args:
            request: an HttpRequest
        """
        super().__init__(*args, **kwargs)
        self.request = request

    def clean_email(self) -> str:
        """Check that `email` has not been blocked.

        Returns:
            the cleaned email

        Raised:
            ValidationError: `email` or its domain is blocked.
        """
        email = self.cleaned_data["email"]
        if BlockList.is_blocked(email):
            raise ValidationError(
                _(
                    "%(email)s has been blocked by Software Heritage and can't be "
                    "used to request an archive alteration. Please contact us if you "
                    "need to unblock this address."
                )
                % {"email": email}
            )
        return email

    def clean(self) -> Optional[dict[str, Any]]:
        """Send the verification email.

        Returns:
            Form's cleaned data.
        """
        cleaned_data = super().clean()
        if cleaned_data:
            email = cleaned_data["email"]
            token = Token.create_for(email)
            send_email_magic_link(token, self.request)
        return cleaned_data


class OriginSearchForm(forms.Form):
    """Search Origins."""

    query = forms.CharField(label=_("Search"), required=True)


class OriginSelectForm(forms.Form):
    """Select Origins."""

    urls = MultipleOriginField(widget=forms.CheckboxSelectMultiple)


# TODO: proper templates
INITIALS_REASONS = {
    "copyright": {
        "reasons": _(
            "The code available in the repos is under the xxx license which does "
            "not allow..."
        ),
        "expected_outcome": _(
            "Example: Please remove archived content for repos ... and prevent future "
            "archiving."
        ),
    },
    "pii": {
        "reasons": _(
            "I've rewritten the history of my repo due to ... your archive still "
            "shows the old content and I need you to delete it."
        ),
        "expected_outcome": _(
            "Example:  Please remove archived content for repos ... and re-archive the "
            "current version."
        ),
    },
    "legal": {
        "reasons": _(
            "malicious content is available on the specified origins [explain what "
            "kind of content]"
        ),
        "expected_outcome": _(
            "Example: I've submitted a takedown request to the legal authorities but "
            "meanwhile please mask the origins so the content is not publicly "
            "available on SWH anymore"
        ),
    },
}


class ReasonsForm(forms.Form):
    """Alteration request's reasons and expected outcome."""

    reasons = forms.CharField(
        label=_("Reasons why Archive content should be changed"),
        help_text=_(
            "Please describe as clearly as possible the reasons for your request."
        ),
        widget=forms.Textarea,
        required=True,
    )
    expected_outcome = forms.CharField(
        label=_("What result are you looking for?"),
        help_text=_(
            "You can specify your expectations regarding the Archive change "
            "mechanisms outlined in the content policy page."
        ),
        widget=forms.Textarea,
        required=True,
    )


class ConfirmationForm(forms.Form):
    """Confirm the alteration request."""

    confirm = forms.BooleanField(
        label=_(
            "I confirm that the information provided in this summary is accurate, "
            "correct, and complete. I attest that this request is made without "
            "unethical or fraudulent intent."
        ),
        required=True,
    )


class CategoryForm(forms.Form):
    """Choose an alteration category.

    This form is solely used for data validation, the assistant_category template will
    display each choices as a submit button.
    """

    category = forms.ChoiceField(
        label=_("Category"),
        choices=AlterationCategory,
        required=True,
    )


class AlterationAccessForm(forms.Form):
    """Security check before accessing an ``Alteration``."""

    email = forms.EmailField(label=_("Your email address"), required=True)

    def __init__(self, *args, alteration: Alteration, request: HttpRequest, **kwargs):
        """Store the extra ``alteration`` & ``request`` parameters.

        Args:
            alteration: an Alteration instance
            request: an HttpRequest
        """
        super().__init__(*args, **kwargs)
        self.alteration = alteration
        self.request = request

    def clean(self) -> Optional[dict[str, Any]]:
        """Check that `email` matches the requested `Alteration`.

        If it matches, send an email containing a magic link to auth the requester, if
        or else do nothing.

        Returns:
            Form's cleaned data.
        """
        cleaned_data = super().clean()
        if cleaned_data and cleaned_data["email"] == self.alteration.email:
            token = Token.create_for(self.alteration)
            send_alteration_magic_link(token, self.request)
        return cleaned_data


class AlterationSearchForm(forms.ModelForm):
    """Search alterations."""

    class Meta:
        model = Alteration
        fields = ["status"]

    query = forms.CharField(label=_("Search"), required=False)
    page = forms.IntegerField(label=_("Page"), required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["status"].required = False

    def search(self) -> Page:
        """Search/filter results and handle pagination.

        Returns:
            A paginated list of `Alteration`.
        """
        if self.cleaned_data.get("query"):
            qs = Alteration.objects.search(self.cleaned_data["query"]).select_related()
        else:
            qs = Alteration.objects.select_related()
        if self.cleaned_data.get("status"):
            qs = qs.filter(status=self.cleaned_data["status"])
        paginator = Paginator(qs, 20)
        return paginator.get_page(self.cleaned_data.get("page", 1))


class OriginAdminForm(forms.ModelForm):
    """Update an ``Origin``."""

    class Meta:
        model = Origin
        exclude = ["id", "alteration"]
        widgets = {
            "code_license": forms.TextInput(),
            "reason": forms.Textarea(attrs={"rows": 4}),
        }

    def __init__(self, *args, request: Optional[HttpRequest] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def save(self, commit=True) -> Origin:
        """Save and generate a changelog if needed."""
        old_values = model_to_dict(self.instance)
        old_url = self.instance.url
        origin = super().save(commit)
        if self.has_changed():
            previous_values = {key: old_values[key] for key in self.changed_data}
            Event.objects.create(
                alteration=self.instance.alteration,
                author=self.request.user.get_username() if self.request else "",
                category=EventCategory.LOG,
                content=generate_origin_changelog(old_url, previous_values),
                internal=False,
            )
        return origin


class AlterationForm(forms.ModelForm):
    """Update an ``Alteration``."""

    class Meta:
        model = Alteration
        exclude = ["id", "origins", "events", "status", "category", "email"]

    def __init__(self, *args, author: str = "", **kwargs):
        super().__init__(*args, **kwargs)
        self.author = author

    def save(self, commit=True) -> Alteration:
        """Save and generate a changelog if needed."""
        old_values = model_to_dict(self.instance)
        alteration = super().save(commit)
        if self.has_changed():
            previous_values = {key: old_values[key] for key in self.changed_data}
            Event.objects.create(
                alteration=alteration,
                author=self.author,
                category=EventCategory.LOG,
                content=generate_alteration_changelog(previous_values),
                internal=False,
            )
        return alteration


class AlterationAdminForm(AlterationForm):
    """Update an ``Alteration``."""

    class Meta:
        model = Alteration
        exclude = ["id", "origins", "events"]


class MessageForm(forms.ModelForm):
    """Message form for requesters."""

    class Meta:
        model = Event
        fields = ["content"]
        widgets = {
            "content": forms.Textarea(attrs={"rows": 4}),
        }
        labels = {"content": _("Your message")}

    def __init__(self, *args, alteration: Alteration, **kwargs):
        super().__init__(*args, **kwargs)
        self.alteration = alteration
        self.fields["content"].required = True

    def save(self, commit=True) -> Event:
        event = super().save(commit=False)
        event.alteration = self.alteration
        event.author = "Requester"
        event.category = EventCategory.MESSAGE
        event.recipient = EventRecipient.SUPPORT
        event.internal = False
        event.save()
        return event


class MessageAdminForm(forms.ModelForm):
    """Message form for admins."""

    class Meta:
        model = Event
        fields = ["recipient", "internal", "content"]
        widgets = {
            "content": forms.Textarea(attrs={"rows": 4}),
        }

    def __init__(self, *args, alteration: Alteration, author: str, **kwargs):
        super().__init__(*args, **kwargs)
        self.alteration = alteration
        self.author = author
        self.fields["recipient"].required = True
        self.fields["content"].required = True

    def clean(self):
        cleaned_data = super().clean()
        recipient = cleaned_data.get("recipient")
        internal = cleaned_data.get("internal")

        if recipient == EventRecipient.REQUESTER and internal:
            self.add_error(
                "internal", _("Can't send an `internal` message to the Requester")
            )
        return cleaned_data

    def save(self, commit=True) -> Event:
        event = super().save(commit=False)
        event.alteration = self.alteration
        event.author = self.author
        event.category = EventCategory.MESSAGE
        event.save()
        return event


class EventAdminForm(forms.ModelForm):
    """Update an ``Event``."""

    class Meta:
        model = Event
        exclude = ["id", "alteration", "category"]
        widgets = {
            "content": forms.Textarea(attrs={"rows": 4}),
        }



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Tuple

from rest_framework.request import Request

from swh.web.save_origin_webhooks.generic_receiver import OriginSaveWebhookReceiver


class BitbucketOriginSaveWebhookReceiver(OriginSaveWebhookReceiver):
    FORGE_TYPE = "Bitbucket"
    WEBHOOK_GUIDE_URL = (
        "https://support.atlassian.com/bitbucket-cloud/docs/manage-webhooks/"
    )
    REPO_TYPES = "git"

    def is_forge_request(self, request: Request) -> bool:
        return (
            request.headers.get("User-Agent", "").startswith(
                f"{self.FORGE_TYPE}-Webhooks/"
            )
            and "X-Event-Key" in request.headers
        )

    def is_push_event(self, request: Request) -> bool:
        return request.headers["X-Event-Key"] == "repo:push"

    def extract_repo_info(self, request: Request) -> Tuple[str, str, bool]:
        repo_url = (
            request.data.get("repository", {})
            .get("links", {})
            .get("html", {})
            .get("href", "")
        )
        if repo_url:
            repo_url += ".git"

        private = request.data.get("repository", {}).get("is_private", False)

        return repo_url, "git", private


api_origin_save_webhook_bitbucket = BitbucketOriginSaveWebhookReceiver()



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.urls import re_path as url

# register Web API endpoints
from swh.web.vault.api_views import vault_api_urls


def vault_view(request: HttpRequest) -> HttpResponse:
    return render(
        request,
        "vault-ui.html",
        {"heading": "Download archive content from the Vault"},
    )


def browse_vault_view(request: HttpRequest) -> HttpResponse:
    return redirect("vault")


urlpatterns = [
    url(r"^vault/$", vault_view, name="vault"),
    # for backward compatibility
    url(r"^browse/vault/$", browse_vault_view, name="browse-vault"),
    *vault_api_urls.get_url_patterns(),
]



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import migrations, models


def _rename_not_yet_scheduled_loading_task_status(apps, schema_editor):
    SaveOriginRequest = apps.get_model("swh_web_save_code_now", "SaveOriginRequest")
    for sor in SaveOriginRequest.objects.all():
        if sor.loading_task_status == "not yet scheduled":
            sor.loading_task_status = "pending"
            sor.save()


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_save_code_now", "0014_saveoriginrequest_snapshot_swhid"),
    ]

    operations = [
        migrations.AlterField(
            model_name="saveoriginrequest",
            name="loading_task_status",
            field=models.TextField(
                choices=[
                    ("not created", "not created"),
                    ("not yet scheduled", "not yet scheduled"),
                    ("scheduled", "scheduled"),
                    ("succeeded", "succeeded"),
                    ("failed", "failed"),
                    ("running", "running"),
                    ("pending", "pending"),
                ],
                default="not created",
            ),
        ),
        migrations.RunPython(_rename_not_yet_scheduled_loading_task_status),
    ]



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.apps import AppConfig

APP_LABEL = "swh_web_alter"


class AlterConfig(AppConfig):
    name = "swh.web.alter"
    label = APP_LABEL



# Copyright (C) 2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.helpers import check_html_get_response
from swh.web.utils import reverse


def test_origin_search_no_js_all_origins_with_snapshot(
    client, origins_with_non_empty_snapshot
):
    url = reverse("browse-search", query_params={"no_js": "on"})
    resp = check_html_get_response(
        client, url, template_used="browse-search.html", status_code=200
    )
    for origin in origins_with_non_empty_snapshot:
        assert_contains(resp, origin.url)



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from base64 import b64encode
from typing import Optional, cast

from pybadges2 import badge

from django.contrib.staticfiles import finders
from django.http import HttpRequest, HttpResponse
from django.middleware.cache import CacheMiddleware

from swh.model.exceptions import ValidationError
from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model.swhids import CoreSWHID, ObjectType, QualifiedSWHID
from swh.web.config import get_config
from swh.web.utils import archive, reverse
from swh.web.utils.exc import BadInputExc, NotFoundExc
from swh.web.utils.identifiers import parse_object_type, resolve_swhid

_orange = "#f36a24"
_blue = "#0172b2"
_red = "#cd5741"

_swh_logo_data = None

badge_config = {
    "content": {
        "color": _blue,
        "title": "Archived source file",
    },
    "directory": {
        "color": _blue,
        "title": "Archived source tree",
    },
    "origin": {
        "color": _orange,
        "title": "Archived software repository",
    },
    "release": {
        "color": _blue,
        "title": "Archived software release",
    },
    "revision": {
        "color": _blue,
        "title": "Archived commit",
    },
    "snapshot": {
        "color": _blue,
        "title": "Archived software repository snapshot",
    },
    "error": {"color": _red, "title": "An error occurred when generating the badge"},
}


def get_logo_data() -> str:
    """
    Get data-URI for Software Heritage SVG logo to embed it in
    the generated badges.
    """
    global _swh_logo_data
    if _swh_logo_data is None:
        swh_logo_path = cast(str, finders.find("img/swh-logo-white.svg"))
        with open(swh_logo_path, "rb") as swh_logo_file:
            _swh_logo_data = "data:image/svg+xml;base64,%s" % b64encode(
                swh_logo_file.read()
            ).decode("ascii")
    return _swh_logo_data


def swh_badge(
    request: HttpRequest,
    object_type: str,
    object_id: str,
    object_swhid: Optional[str] = "",
) -> HttpResponse:
    """
    Generate a Software Heritage badge for a given object type and id.

    Args:
        request: input http request
        object_type: The type of swh object to generate a badge for,
            either *content*, *directory*, *revision*, *release*, *origin*
            or *snapshot*
        object_id: The id of the swh object, either an url for origin
            type or a *sha1* for other object types
        object_swhid: If provided, the object SWHID will not be recomputed

    Returns:
        HTTP response with content type *image/svg+xml* containing the SVG
        badge data. If the provided parameters are invalid, HTTP 400 status
        code will be returned. If the object cannot be found in the archive,
        HTTP 404 status code will be returned.

    """
    left_text = "error"
    whole_link: str | None = None
    status_code = 200

    # cache badge for 30 days by default
    cache_timeout = (
        get_config().get("badges", {}).get("cache_timeout", 30 * 24 * 60 * 60)
    )

    def get_response(request: HttpRequest) -> HttpResponse:
        return HttpResponse()

    cachemiddleware = CacheMiddleware(
        cache_timeout=int(cache_timeout),
        get_response=get_response,
    )

    if (response := cachemiddleware.process_request(request)) is not None:
        # return cached badge response
        return cast(HttpResponse, response)

    setattr(request, "_cache_update_cache", False)

    try:
        if object_type == "origin":
            archive.lookup_origin(object_id)
            right_text = "repository"
            whole_link = reverse(
                "browse-origin", query_params={"origin_url": object_id}
            )
        else:
            # when SWHID is provided, object type and id will be parsed
            # from it
            if object_swhid:
                parsed_swhid = QualifiedSWHID.from_string(object_swhid)
                parsed_object_type = parsed_swhid.object_type
                object_id = hash_to_hex(parsed_swhid.object_id)
                swh_object = archive.lookup_object(parsed_swhid.object_type, object_id)
                # remove SWHID qualified if any for badge text
                right_text = str(
                    CoreSWHID(
                        object_type=parsed_swhid.object_type,
                        object_id=parsed_swhid.object_id,
                    )
                )
                object_type = parsed_swhid.object_type.name.lower()
            else:
                parsed_object_type = parse_object_type(object_type)
                right_text = str(
                    CoreSWHID(
                        object_type=parsed_object_type,
                        object_id=hash_to_bytes(object_id),
                    )
                )
                swh_object = archive.lookup_object(parsed_object_type, object_id)

            whole_link = resolve_swhid(str(right_text))["browse_url"]
            # use release name for badge text
            if parsed_object_type == ObjectType.RELEASE:
                right_text = "release %s" % swh_object["name"]
        left_text = "archived"
        # cache badge response when targeted object is found
        setattr(request, "_cache_update_cache", True)
    except (BadInputExc, ValidationError, ValueError):
        right_text = f'invalid {object_type if object_type else "object"} id'
        object_type = "error"
        status_code = 400
    except NotFoundExc:
        right_text = f'{object_type if object_type else "object"} not found'
        object_type = "error"
        status_code = 404

    badge_data = badge(
        left_text=left_text,
        right_text=right_text,
        right_color=badge_config[object_type]["color"],
        whole_link=request.build_absolute_uri(whole_link),
        whole_title=badge_config[object_type]["title"],
        logo=get_logo_data(),
        embed_logo=True,
    )

    response = HttpResponse(
        badge_data, content_type="image/svg+xml", status=status_code
    )
    return cast(HttpResponse, cachemiddleware.process_response(request, response))


def swh_badge_swhid(request: HttpRequest, object_swhid: str) -> HttpResponse:
    """
    Generate a Software Heritage badge for a given object SWHID.

    Args:
        request (django.http.HttpRequest): input http request
        object_swhid (str): a SWHID of an archived object

    Returns:
        django.http.HttpResponse: An http response with content type
            *image/svg+xml* containing the SVG badge data. If any error
            occurs, a status code of 400 will be returned.
    """
    return swh_badge(request, "", "", object_swhid)



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.contrib.auth.views import LoginView, LogoutView
from django.http import HttpRequest
from django.http.response import HttpResponse
from django.shortcuts import render
from django.urls import re_path as url

from swh.auth.django.views import urlpatterns as auth_urlpatterns
from swh.web.auth.views import (
    oidc_generate_bearer_token,
    oidc_generate_bearer_token_complete,
    oidc_get_bearer_token,
    oidc_list_bearer_tokens,
    oidc_profile_view,
    oidc_revoke_bearer_tokens,
)
from swh.web.config import get_config, oidc_enabled

config = get_config()
urlpatterns = []

if not oidc_enabled():
    urlpatterns = [
        url(
            r"^login/$",
            LoginView.as_view(template_name="login.html"),
            name="login",
        )
    ]

if oidc_enabled() or config["e2e_tests_mode"]:
    urlpatterns += auth_urlpatterns + [
        url(
            r"^oidc/generate-bearer-token/$",
            oidc_generate_bearer_token,
            name="oidc-generate-bearer-token",
        ),
        url(
            r"^oidc/generate-bearer-token-complete/$",
            oidc_generate_bearer_token_complete,
            name="oidc-generate-bearer-token-complete",
        ),
        url(
            r"^oidc/list-bearer-token/$",
            oidc_list_bearer_tokens,
            name="oidc-list-bearer-tokens",
        ),
        url(
            r"^oidc/get-bearer-token/$",
            oidc_get_bearer_token,
            name="oidc-get-bearer-token",
        ),
        url(
            r"^oidc/revoke-bearer-tokens/$",
            oidc_revoke_bearer_tokens,
            name="oidc-revoke-bearer-tokens",
        ),
        url(
            r"^oidc/profile/$",
            oidc_profile_view,
            name="oidc-profile",
        ),
    ]


def logout_page(request: HttpRequest) -> HttpResponse:
    return render(request, template_name="logout.html")


urlpatterns += [
    url(
        r"^logout/$",
        LogoutView.as_view(template_name="logout.html"),
        name="logout",
    ),
    url(
        r"^logout-page/$",
        logout_page,
        name="logout-page",
    ),
]



# Copyright (C) 2019-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
import json

import pytest

from swh.auth.django.utils import oidc_user_from_profile
from swh.web.save_code_now.models import SaveOriginRequest
from swh.web.save_code_now.origin_save import SAVE_REQUEST_ACCEPTED, SAVE_TASK_SUCCEEDED
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.helpers import check_html_post_response, check_http_get_response
from swh.web.utils import reverse

VISIT_TYPES = ("git", "svn", "hg", "cvs", "bzr")
PRIVILEGED_VISIT_TYPES = tuple(list(VISIT_TYPES) + ["archives"])


def test_old_save_url_redirection(client):
    url = reverse("browse-origin-save")
    redirect_url = reverse("origin-save")

    resp = check_http_get_response(client, url, status_code=302)
    assert resp["location"] == redirect_url


def test_origin_save_form_autofill(client):
    origin_url = "https://git.example.org/project"
    url = reverse("origin-save", query_params={"origin_url": origin_url})
    resp = check_http_get_response(client, url, status_code=200)
    assert_contains(resp, origin_url)


@pytest.mark.django_db
def test_save_origin_requests_list(client, mocker):
    nb_origins_per_type = 10
    for visit_type in VISIT_TYPES:
        for i in range(nb_origins_per_type):
            SaveOriginRequest.objects.create(
                request_date=datetime.now(tz=timezone.utc),
                visit_type=visit_type,
                origin_url=f"https://{visit_type}.example.org/project{i}",
                status=SAVE_REQUEST_ACCEPTED,
                visit_date=datetime.now(tz=timezone.utc) + timedelta(hours=1),
                loading_task_id=i,
                loading_task_status=SAVE_TASK_SUCCEEDED,
            )

    mock_scheduler = mocker.patch("swh.web.save_code_now.origin_save.scheduler")
    mock_scheduler.get_tasks.return_value = []
    mock_scheduler.get_task_runs.return_value = []

    # retrieve all save requests in 3 pages, sorted in descending order
    # of request creation
    for i, visit_type in enumerate(reversed(VISIT_TYPES)):
        url = reverse(
            "origin-save-requests-list",
            url_args={"status": "all"},
            query_params={
                "draw": i + 1,
                "search[value]": "",
                "order[0][column]": "0",
                "order[0][name]": "request_date",
                "columns[0][name]": "request_date",
                "order[0][dir]": "desc",
                "length": nb_origins_per_type,
                "start": i * nb_origins_per_type,
            },
        )

        resp = check_http_get_response(
            client, url, status_code=200, content_type="application/json"
        )
        sors = json.loads(resp.content.decode("utf-8"))
        assert sors["draw"] == i + 1
        assert sors["recordsFiltered"] == len(VISIT_TYPES) * nb_origins_per_type
        assert sors["recordsTotal"] == len(VISIT_TYPES) * nb_origins_per_type
        assert len(sors["data"]) == nb_origins_per_type
        assert all(d["visit_type"] == visit_type for d in sors["data"])

    # retrieve save requests filtered by visit type in a single page
    for i, visit_type in enumerate(reversed(VISIT_TYPES)):
        url = reverse(
            "origin-save-requests-list",
            url_args={"status": "all"},
            query_params={
                "draw": i + 1,
                "search[value]": visit_type,
                "order[0][column]": "0",
                "order[0][name]": "request_date",
                "columns[0][name]": "request_date",
                "order[0][dir]": "desc",
                "length": nb_origins_per_type,
                "start": 0,
            },
        )

        resp = check_http_get_response(
            client, url, status_code=200, content_type="application/json"
        )
        sors = json.loads(resp.content.decode("utf-8"))
        assert sors["draw"] == i + 1
        assert sors["recordsFiltered"] == nb_origins_per_type
        assert sors["recordsTotal"] == len(VISIT_TYPES) * nb_origins_per_type
        assert len(sors["data"]) == nb_origins_per_type
        assert all(d["visit_type"] == visit_type for d in sors["data"])


@pytest.mark.django_db
def test_save_origin_requests_list_user_filter(client, mocker, keycloak_oidc):
    # anonymous user created a save request
    sor = SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type="svn",
        origin_url="https://svn.example.org/user/project",
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=datetime.now(tz=timezone.utc) + timedelta(hours=1),
        loading_task_id=1,
        loading_task_status=SAVE_TASK_SUCCEEDED,
    )

    # authenticated user created a save request
    user = oidc_user_from_profile(keycloak_oidc, keycloak_oidc.login())
    client.login(code="", code_verifier="", redirect_uri="")

    sor = SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type="git",
        origin_url="https://git.example.org/user/project",
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=datetime.now(tz=timezone.utc) + timedelta(hours=1),
        loading_task_id=2,
        loading_task_status=SAVE_TASK_SUCCEEDED,
        user_ids=f'"{user.id}"',
    )

    # filter save requests according to user id
    url = reverse(
        "origin-save-requests-list",
        url_args={"status": "all"},
        query_params={
            "draw": 1,
            "search[value]": "",
            "order[0][column]": "0",
            "order[0][name]": "request_date",
            "columns[0][name]": "request_date",
            "order[0][dir]": "desc",
            "length": 10,
            "start": "0",
            "user_requests_only": "1",
        },
    )

    resp = check_http_get_response(
        client, url, status_code=200, content_type="application/json"
    )
    sors = json.loads(resp.content.decode("utf-8"))
    assert sors["recordsFiltered"] == 1
    assert sors["recordsTotal"] == 2
    assert sors["data"][0] == sor.to_dict()


def test_admin_origin_save_requests_csv_dump_anonymous(client):
    url = reverse("admin-origin-save-requests-csv")
    login_url = reverse("oidc-login", query_params={"next": url})
    resp = check_http_get_response(client, url, status_code=302)
    assert resp["location"] == login_url


@pytest.mark.django_db
def test_admin_origin_save_requests_csv_dump_staff(client, staff_user):
    client.force_login(staff_user)
    url = reverse("admin-origin-save-requests-csv")
    check_http_get_response(client, url, status_code=200)


@pytest.mark.django_db
def test_save_origin_requests_list_without_query_params(client):
    check_http_get_response(
        client,
        reverse("origin-save-requests-list", url_args={"status": "all"}),
        status_code=200,
        content_type="application/json",
    )


@pytest.mark.django_db
def test_admin_origin_save_authorized_urls_without_query_params(client, staff_user):
    client.force_login(staff_user)
    url = reverse("admin-origin-save-authorized-urls-list")
    check_http_get_response(client, url, status_code=200)


@pytest.mark.django_db
@pytest.mark.parametrize(
    "anonymous", [False, True], ids=["anonymous user", "staff user"]
)
def test_origin_save_view_create_save_request(
    client, staff_user, swh_scheduler, anonymous
):
    origin_url = "https://git.example.org/user/test_project"
    if not anonymous:
        client.force_login(staff_user)
    url = reverse("origin-save")
    resp = check_html_post_response(
        client,
        url,
        data={"origin_url": origin_url, "visit_type": "git"},
        status_code=200,
        template_used="origin-save-help.html",
    )
    if not anonymous:
        assert_contains(resp, 'The "save code now" request has been accepted')
    else:
        assert_contains(
            resp, 'The "save code now" request has been put in pending state'
        )



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

import pytest


@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(args, early_config, parser):
    # avoid side effects in tests if a local config file exists
    os.environ["SWH_CONFIG_FILENAME"] = "non/existing/config"



# Copyright (C) 2015-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import traceback
from typing import Optional, TypeVar

from django_ratelimit.exceptions import Ratelimited
import sentry_sdk

from django.core import exceptions
from django.http import Http404, HttpRequest, HttpResponse
from django.http.response import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotFound,
    HttpResponseServerError,
    JsonResponse,
)
from django.shortcuts import render
from django.template import loader
from django.utils.html import escape, format_html
from rest_framework.exceptions import APIException
from rest_framework.renderers import JSONRenderer

from swh.core.api import RemoteException, TransientRemoteException
from swh.storage.exc import MaskedObjectException
from swh.web.api.renderers import YAMLRenderer
from swh.web.config import get_config

logger = logging.getLogger("django")


class BadInputExc(ValueError):
    """Wrong request to the api.

    Example: Asking a content with the wrong identifier format.

    """

    pass


class NotFoundExc(exceptions.ObjectDoesNotExist):
    """Good request to the api but no results were found.

    Example: Asking a content with the right identifier format but
    that content does not exist.

    """

    def __str__(self):
        return self.args[0]


class ForbiddenExc(exceptions.PermissionDenied):
    """Good request to the api, forbidden result to return due to enforce
       policy.

    Example: Asking for a raw content which exists but whose mimetype
    is not text.

    """

    pass


class UnauthorizedExc(Exception):
    """Request to Web API endpoint requires authentication."""

    pass


class LargePayloadExc(Exception):
    """The input size is too large.

    Example: Asking to resolve 10000 SWHIDs when the limit is 1000.
    """

    pass


http_status_code_message = {
    400: "Bad Request",
    401: "Unauthorized",
    403: "Access Denied",
    404: "Resource not found",
    413: "Payload Too Large",
    429: "Too Many Requests",
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service unavailable",
}

HttpResponseType = TypeVar("HttpResponseType", bound=HttpResponse)


def _generate_error_page(
    request: HttpRequest,
    error_description: str,
    http_response: type[HttpResponseType],
) -> HttpResponseType:
    error_code = http_response.status_code
    error_data = {
        "error": http_status_code_message[error_code],
        "reason": error_description,
    }

    accepted_media_type = request.headers.get("Accept", "application/json")

    # ensure django error is returned using same content type as Accept request header
    if accepted_media_type in ("application/json", "*/*"):
        return http_response(
            JSONRenderer().render(error_data),
            content_type="application/json",
        )
    elif accepted_media_type == "application/yaml":
        return http_response(
            YAMLRenderer().render(error_data),
            content_type="application/yaml",
        )
    else:
        return http_response(
            loader.render_to_string(
                template_name="error.html",
                context={
                    "error_code": error_code,
                    "error_message": http_status_code_message[error_code],
                    "error_description": error_description,
                },
                request=request,
            ),
            content_type="text/html",
        )


def masked_to_common_types(exc: MaskedObjectException):
    """Convert ``exc.masked`` to common types, suitable for
    JSON and YAML encoding.

    ExtendedSWHID becomes strings and MaskedStatus becomes a dict.
    """
    return {
        str(swhid): [
            {
                "request": status.request,
                "status": status.state.name.lower().replace("_", "-"),
            }
            for status in statuses
        ]
        for swhid, statuses in exc.masked.items()
    }


def _generate_masked_object_page(
    request: HttpRequest, exc: MaskedObjectException
) -> HttpResponse:
    error_code = 403  # Forbidden
    error_data = {
        "error": http_status_code_message[error_code],
        "reason": str(exc),
        "masked": masked_to_common_types(exc),
    }

    accepted_media_type = request.headers.get("Accept", "application/json")

    if accepted_media_type in ("application/json", "*/*"):
        return JsonResponse(
            error_data,
            status=error_code,
        )
    elif accepted_media_type == "application/yaml":
        return HttpResponse(
            YAMLRenderer().render(error_data),
            content_type="application/yaml",
            status=error_code,
        )
    else:
        return render(
            request,
            "masked.html",
            {
                "error_code": error_code,
                "error_message": "Access restricted",
                "error_description": str(exc),
                "masked": exc.masked,
            },
            status=error_code,
        )


def swh_handle400(
    request: HttpRequest,
    exception: Optional[Exception] = None,
    template_name: str = "",
) -> HttpResponseBadRequest:
    """
    Custom Django HTTP error 400 handler for swh-web.
    """
    error_description = (
        "The server cannot process the request to %s due to "
        "something that is perceived to be a client error."
        % escape(request.META["PATH_INFO"])
    )
    return _generate_error_page(request, error_description, HttpResponseBadRequest)


def swh_handle403(
    request,
    exception: Optional[Exception] = None,
    template_name: str = "",
) -> HttpResponseForbidden:
    """
    Custom Django HTTP error 403 handler for swh-web.
    """
    error_description = "The resource %s requires an authentication." % escape(
        request.META["PATH_INFO"]
    )
    return _generate_error_page(request, error_description, HttpResponseForbidden)


def swh_handle404(
    request,
    exception: Optional[Exception] = None,
    template_name: str = "",
) -> HttpResponseNotFound:
    """
    Custom Django HTTP error 404 handler for swh-web.
    """
    error_description = "The resource %s could not be found on the server." % escape(
        request.META["PATH_INFO"]
    )
    return _generate_error_page(request, error_description, HttpResponseNotFound)


def swh_handle500(
    request: HttpRequest,
    template_name: str = "",
) -> HttpResponseServerError:
    """
    Custom Django HTTP error 500 handler for swh-web.
    """
    error_description = (
        "An unexpected condition was encountered when "
        "requesting resource %s." % escape(request.META["PATH_INFO"])
    )
    return _generate_error_page(request, error_description, HttpResponseServerError)


def sentry_capture_exception(exc: Exception) -> None:
    if isinstance(
        exc,
        (
            exceptions.ObjectDoesNotExist,
            exceptions.DisallowedHost,
            exceptions.PermissionDenied,
            BadInputExc,
            NotFoundExc,
            RemoteException,
            TransientRemoteException,
            Ratelimited,
            UnauthorizedExc,
        ),
    ):
        # ignore noisy exceptions we cannot do anything about
        pass
    elif isinstance(exc, APIException) and 400 <= exc.status_code < 500:
        # ignore client errors (4xx status codes)
        pass
    else:
        # log everything else
        sentry_sdk.capture_exception(exc)


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401


class HttpResponseTooManyRequests(HttpResponse):
    status_code = 429


def handle_view_exception(request: HttpRequest, exc: Exception) -> HttpResponse:
    """
    Function used to generate an error page when an exception
    was raised inside a swh-web browse view.
    """
    sentry_capture_exception(exc)

    http_response: type[HttpResponse] = HttpResponseServerError
    if isinstance(exc, BadInputExc):
        http_response = HttpResponseBadRequest
    elif isinstance(exc, UnauthorizedExc):
        http_response = HttpResponseUnauthorized
    elif isinstance(exc, Ratelimited):
        http_response = HttpResponseTooManyRequests
    elif isinstance(exc, exceptions.PermissionDenied):
        http_response = HttpResponseForbidden
    elif isinstance(exc, (Http404, NotFoundExc)):
        http_response = HttpResponseNotFound
    elif isinstance(exc, MaskedObjectException):
        return _generate_masked_object_page(request, exc)

    error_description = format_html("{}: {}", type(exc).__name__, str(exc))
    if get_config()["debug"]:
        traceback_str = traceback.format_exc()
        logger.error(traceback_str)
        error_description = escape(traceback_str)
    elif http_response.status_code == 500:
        logger.exception(exc)

    resp = _generate_error_page(request, error_description, http_response)
    if get_config()["debug"]:
        resp.traceback = error_description  # type: ignore[attr-defined]
    return resp



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import uuid

from django.db import migrations, models
import django.db.models.deletion

import swh.web.alter.models


def _create_alter_permission(apps, schema_editor):
    from swh.web.auth.utils import (
        ALTER_ADMIN_PERMISSION,
        get_or_create_django_permission,
    )

    get_or_create_django_permission(ALTER_ADMIN_PERMISSION)


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Alteration",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated"),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            (None, "Filter by status"),
                            ("validating", "Validating"),
                            ("planning", "Planning"),
                            ("executing", "Executing"),
                            ("processed", "Processed"),
                            ("rejected", "Rejected"),
                            ("closed", "Closed"),
                            ("archived", "Archived"),
                        ],
                        default="validating",
                        max_length=20,
                        verbose_name="status",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[
                            (None, "Filter by category"),
                            ("copyright", "Copyright / License infringement"),
                            ("pii", "Personal Identifiable Information"),
                            ("legal", "Other legal matters"),
                        ],
                        max_length=20,
                        verbose_name="category",
                    ),
                ),
                ("reasons", models.TextField(verbose_name="reasons")),
                ("expected_outcome", models.TextField(verbose_name="expected outcome")),
                (
                    "email",
                    models.EmailField(max_length=254, verbose_name="requester's email"),
                ),
            ],
            options={
                "db_table": "alteration",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="BlockList",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated"),
                ),
                (
                    "email_or_domain",
                    swh.web.alter.models.LowerCharField(
                        max_length=254,
                        validators=[swh.web.alter.models.validate_email_or_domain],
                        verbose_name="email or domain",
                    ),
                ),
                ("reason", models.TextField(blank=True, verbose_name="reasons")),
            ],
            options={
                "db_table": "block_list",
                "indexes": [
                    models.Index(
                        fields=["email_or_domain"], name="block_list_email_o_123185_idx"
                    )
                ],
                "constraints": [
                    models.UniqueConstraint(
                        fields=("email_or_domain",), name="unique_email_or_domain"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated"),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[("log", "Event"), ("message", "Message")],
                        max_length=20,
                        verbose_name="category",
                    ),
                ),
                (
                    "author",
                    models.CharField(blank=True, max_length=200, verbose_name="author"),
                ),
                (
                    "recipient",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("requester", "Requester"),
                            ("support", "Support"),
                            ("manager", "Manager"),
                            ("legal", "Legal"),
                            ("technical", "Technical"),
                        ],
                        max_length=20,
                        verbose_name="recipient role",
                    ),
                ),
                ("content", models.TextField(blank=True, verbose_name="content")),
                (
                    "internal",
                    models.BooleanField(
                        default=True,
                        help_text=(
                            "Internal messages are not visible in the Requester "
                            "activity log to avoid unnecessary noise, they must not "
                            "be used for confidential exchanges between team members "
                            "and could be requested by the user."
                        ),
                        verbose_name="internal",
                    ),
                ),
                (
                    "alteration",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.RESTRICT,
                        related_name="events",
                        to="swh_web_alter.alteration",
                    ),
                ),
            ],
            options={
                "db_table": "event",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="Origin",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated"),
                ),
                ("url", models.URLField(verbose_name="URL")),
                (
                    "outcome",
                    models.CharField(
                        choices=[
                            ("validating", "Validating"),
                            ("rejected", "Rejected"),
                            ("mailmap", "Mailmap"),
                            ("mask", "Permanent mask"),
                            ("takedown", "Takedown"),
                            ("block", "Takedown and block"),
                        ],
                        default="validating",
                        max_length=20,
                        verbose_name="outcome",
                    ),
                ),
                (
                    "reason",
                    models.TextField(
                        blank=True, verbose_name="reason for this outcome"
                    ),
                ),
                (
                    "code_license",
                    models.TextField(blank=True, verbose_name="license found in code"),
                ),
                (
                    "available",
                    models.BooleanField(
                        help_text="Is this URL is still available online ?",
                        null=True,
                        verbose_name="available",
                    ),
                ),
                (
                    "ownership",
                    models.CharField(
                        choices=[
                            ("unknown", "?"),
                            ("owner", "Requester is the owner of the origin"),
                            ("fork", "Fork of an origin owned by the requester"),
                            ("other", "Origin has no direct link with the requester"),
                        ],
                        default="unknown",
                        help_text="Is Requester the owner of this origin or is it a fork ?",
                        max_length=20,
                        verbose_name="owner",
                    ),
                ),
                (
                    "alteration",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.RESTRICT,
                        related_name="origins",
                        to="swh_web_alter.alteration",
                    ),
                ),
            ],
            options={
                "db_table": "origin",
                "indexes": [models.Index(fields=["url"], name="origin_url_4c8f23_idx")],
                "constraints": [
                    models.UniqueConstraint(
                        fields=("alteration_id", "url"), name="unique_url"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="Token",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated"),
                ),
                (
                    "email",
                    models.EmailField(max_length=254, null=True, verbose_name="email"),
                ),
                (
                    "value",
                    models.CharField(
                        default=swh.web.alter.models._default_token_value,
                        max_length=32,
                        verbose_name="value",
                    ),
                ),
                (
                    "expires_at",
                    models.DateTimeField(
                        default=swh.web.alter.models._default_token_expires_at,
                        verbose_name="expiration date",
                    ),
                ),
                (
                    "alteration",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tokens",
                        to="swh_web_alter.alteration",
                    ),
                ),
            ],
            options={
                "db_table": "token",
                "indexes": [
                    models.Index(fields=["value"], name="token_value_feb983_idx"),
                    models.Index(fields=["email"], name="token_email_715e44_idx"),
                ],
                "constraints": [
                    models.UniqueConstraint(
                        fields=("value",), name="unique_token_value"
                    )
                ],
            },
        ),
        migrations.RunPython(_create_alter_permission, migrations.RunPython.noop),
    ]



# Copyright (C) 2024-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information
from rest_framework import serializers
from rest_framework.request import Request

from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.utils import enrich_extid
from swh.web.api.views.utils import api_lookup
from swh.web.utils import archive

EXTID_FORMATS = [
    ("hex", "Hexadecimal"),
    ("base64url", "Base64 URL"),
    ("raw", "Raw"),
]


class ExtidTargetQuerySerializer(serializers.Serializer):
    """Extid Target query parameters serializer."""

    extid_type = serializers.CharField(required=False, max_length=255, default=None)
    extid_version = serializers.IntegerField(required=False, min_value=0, default=None)
    extid_format = serializers.ChoiceField(
        required=False, choices=EXTID_FORMATS, default="hex"
    )

    def validate(self, data: dict) -> dict:
        """Validate extid_version and extid_type.

        extid_version is required if extid_type is defined and vice versa.

        Args:
            data: query parameters

        Raises:
            serializers.ValidationError: invalid parameters

        Returns:
            query parameters
        """
        extid_version = data["extid_version"]
        extid_type = data["extid_type"]
        if (extid_version is not None and extid_type is None) or (
            extid_version is None and extid_type is not None
        ):
            raise serializers.ValidationError(
                "Both extid_type and extid_version query parameters must be provided"
            )
        return data


@api_route(
    "/extid/target/<swhid:swhid>/",
    "api-1-extid-target",
    query_params_serializer=ExtidTargetQuerySerializer,
)
@api_doc("/extid/target/", category="External IDentifiers")
@format_docstring()
def api_extid_target(request: Request, swhid: str, validated_query_params: dict):
    """
    .. http:get:: /api/1/extid/target/(swhid)/

        Get information about external identifiers targeting an archived object.

        An `external identifier
        <https://docs.softwareheritage.org/devel/glossary.html#term-external-identifier>`_
        is used by a system that does not fit the Software Heritage data model.

        :param string swhid: a SWHID to check if external identifiers target it
        :query string extid_type: optional external identifier type to use as a filter,
            must be provided if ``extid_version`` parameter is.
        :query number extid_version: optional version number of external identifier
            type, must be provided if ``extid_type`` parameter is.
        :query string extid_format: the format used to encode an extid to an ASCII
            string, either ``base64url``, ``hex`` or ``raw`` (default to ``hex``).

        {common_headers}

        :>jsonarr string extid_type: the type of the external identifier
        :>jsonarr string extid: the value of the external identifier
        :>jsonarr string target: the SWHID of the archived object targeted by the
            external identifier
        :>jsonarr string target_url: URL to browse the targeted archived object
        :>jsonarr number extid_version: the version of the external identifier

        :statuscode 200: no error
        :statuscode 400: provided parameters are invalid
        :statuscode 404: external identifier targeting SWHID cannot be found

        **Example:**

        .. parsed-literal::

            :swh_web_api:`extid/target/swh:1:rev:a2903689803b2c07890a930284425838436425a6/?extid_format=raw`
            :swh_web_api:`extid/target/swh:1:rev:6b29add7cb6b5f6045df308c43e4177f1f854a56/?extid_format=hex`

    """

    extid_type = validated_query_params["extid_type"]
    extid_version = validated_query_params["extid_version"]
    extid_format = validated_query_params["extid_format"]

    return api_lookup(
        archive.lookup_extid_by_target,
        swhid,
        extid_type=extid_type,
        extid_version=extid_version,
        extid_format=extid_format,
        enrich_fn=enrich_extid,
        request=request,
    )


@api_route(
    r"/extid/(?P<extid_type>.+?)/(?P<extid_format>.+?):(?P<extid>.+)/",
    "api-1-extid",
)
@api_doc("/extid/", category="External IDentifiers")
@format_docstring(hg_template="'{node}'")
def api_extid(request: Request, extid_type: str, extid_format: str, extid: str):
    """
    .. http:get:: /api/1/extid/(extid_type)/(extid_format):(extid)/

        Get information about an external identifier.

        An `external identifier
        <https://docs.softwareheritage.org/devel/glossary.html#term-external-identifier>`_
        is used by a system that does not fit the Software Heritage data model.

        As an external identifier is stored in binary into the archive database, the format
        used to decode its ASCII representation must be explicitly specified. The available
        formats are the following:

        - ``base64url``: the external identifier is encoded to `base64url
          <https://datatracker.ietf.org/doc/html/rfc4648#section-5>`_.
        - ``hex``: the external identifier is a checksum in hexadecimal representation
        - ``raw``: the external identifier is an ASCII string

        The types of external identifier that can be requested are given below.

        VCS related:

        - ``bzr-nodeid``: Revision ASCII identifier of a Bazaar repository, to get such
          identifiers use the following command in your Bazaar repository:
          ``bzr log --show-ids``.

        - ``hg-nodeid``: Node hash identifier for the revision of a Mercurial repository,
          to get such identifier execute the following command in your Mercurial repository:
          ``hg id -r <rev_num> --template {hg_template}``.

        `Guix <https://guix.gnu.org/manual/en/html_node/Invoking-guix-hash.html>`_ and
        `Nix <https://nixos.wiki/wiki/Nix_Hash>`_ related (must be queried with the
        **extid_version** query parameter set to **1** to ensure correctness):

        - ``nar-sha256``: sha256 checksum of a `Nix Archive (NAR)
          <https://edolstra.github.io/pubs/phd-thesis.pdf#page=101>`_, used to deterministically
          identifies the contents of a source tree (corresponds to *recursive* hash mode used
          by Guix and Nix)

        - ``checksum-sha256``: sha256 checksum of a file, typically a tarball (corresponds to
          *flat* hash mode used by Guix and Nix)

        - ``checksum-sha512``: sha512 checksum of a file, typically a tarball (corresponds to
          *flat* hash mode used by Guix and Nix)

        :param string extid_type: the type of external identifier
        :param string extid_format: the format used to encode the extid to an ASCII string,
            either ``base64url``, ``hex`` or ``raw``
        :param string extid: the external identifier value
        :query number extid_version: optional version number of external identifier type

        {common_headers}

        :>json string extid_type: the type of the external identifier
        :>json string extid: the value of the external identifier
        :>json string target: the SWHID of the archived object targeted by the
            external identifier
        :>json string target_url: URL to browse the targeted archived object
        :>json number extid_version: the version of the external identifier

        :statuscode 200: no error
        :statuscode 404: requested external identifier cannot be found

        **Example:**

        .. parsed-literal::

            :swh_web_api:`extid/bzr-nodeid/raw:rodney.dawes@canonical.com-20090512192901-f22ja60nsgq9j5a4/`
            :swh_web_api:`extid/hg-nodeid/hex:1ce49c60732c9020ce2f98d03a7a71ec8d5be191/`
            :swh_web_api:`extid/checksum-sha256/base64url:s4lFKlaGmGiN2jiAIGg3ihbBXEr5sVPN2ZtlORKSu8c/?extid_version=1`
            :swh_web_api:`extid/nar-sha256/base64url:AAAlhKVqm86FeTUVYEKY-LOx6Ul-APxjYaDC5zHAY_M/?extid_version=1`
            :swh_web_api:`extid/checksum-sha512/base64url:AL5bxZ-gStT5UpzSc1dN-XVxxWN9FHtvBlZoFeFFMowwgMKWq9GLZHV8DWX-g7ugiKxlKa2ph2oTQCqvhixDQw/?extid_version=1`

    """

    extid_version = request.GET.get("extid_version")

    return api_lookup(
        archive.lookup_extid,
        extid_type,
        extid_format,
        extid,
        extid_version=int(extid_version) if extid_version else None,
        enrich_fn=enrich_extid,
        request=request,
    )



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.urls import path as url

from swh.web.badges import swh_badge, swh_badge_swhid
from swh.web.utils.url_path_converters import register_url_path_converters

register_url_path_converters()

urlpatterns = [
    url(
        "badge/<swhid:object_swhid>/",
        swh_badge_swhid,
        name="swh-badge-swhid",
    ),
    url(
        "badge/<str:object_type>/<path:object_id>/",
        swh_badge,
        name="swh-badge",
    ),
]



# Generated by Django 2.2.28 on 2022-08-16 13:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_add_forge_now", "0007_rename_denied_request_status"),
    ]

    operations = [
        migrations.AlterField(
            model_name="request",
            name="forge_url",
            field=models.URLField(),
        ),
    ]



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List, Optional, Tuple, Union

from django.http import HttpRequest

from swh.model.model import Origin
from swh.web.utils import resolve_branch_alias, reverse
from swh.web.utils.query import parse_hash
from swh.web.utils.typing import OriginInfo


def filter_field_keys(data, field_keys):
    """Given an object instance (directory or list), and a csv field keys
    to filter on.

    Return the object instance with filtered keys.

    Note: Returns obj as is if it's an instance of types not in (dictionary,
    list)

    Args:
        - data: one object (dictionary, list...) to filter.
        - field_keys: csv or set of keys to filter the object on

    Returns:
        obj filtered on field_keys

    """
    if isinstance(data, map):
        return map(lambda x: filter_field_keys(x, field_keys), data)
    if isinstance(data, list):
        return [filter_field_keys(x, field_keys) for x in data]
    if isinstance(data, dict):
        return {k: v for (k, v) in data.items() if k in field_keys}
    return data


def person_to_string(person):
    """Map a person (person, committer, tagger, etc...) to a string."""
    return "".join([person["name"], " <", person["email"], ">"])


def enrich_object(
    object: Dict[str, str], request: Optional[HttpRequest] = None
) -> Dict[str, str]:
    """Enrich an object (revision, release) with link to the 'target' of
    type 'target_type'.

    Args:
        object: An object with target and target_type keys
        (e.g. release, revision)
        request: Absolute URIs will be generated if provided

    Returns:
        Object enriched with target object url (revision, release, content,
        directory)

    """
    if "target" in object and "target_type" in object:
        if object["target_type"] in ("revision", "release", "directory"):
            object["target_url"] = reverse(
                "api-1-%s" % object["target_type"],
                url_args={"sha1_git": object["target"]},
                request=request,
            )
        elif object["target_type"] == "content":
            object["target_url"] = reverse(
                "api-1-content",
                url_args={"q": "sha1_git:" + object["target"]},
                request=request,
            )
        elif object["target_type"] == "snapshot":
            object["target_url"] = reverse(
                "api-1-snapshot",
                url_args={"snapshot_id": object["target"]},
                request=request,
            )

    return object


enrich_release = enrich_object


def enrich_directory_entry(
    directory: Dict[str, str], request: Optional[HttpRequest] = None
) -> Dict[str, str]:
    """Enrich directory entry with url to target.

    Args:
        directory: dict of data associated to a swh directory entry
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched directory dict filled with additional url
    """
    if "type" in directory:
        target_type = directory["type"]
        target = directory["target"]
        if target_type == "file":
            directory["target_url"] = reverse(
                "api-1-content", url_args={"q": "sha1_git:%s" % target}, request=request
            )
        elif target_type == "dir":
            directory["target_url"] = reverse(
                "api-1-directory", url_args={"sha1_git": target}, request=request
            )
        else:
            directory["target_url"] = reverse(
                "api-1-revision", url_args={"sha1_git": target}, request=request
            )

    return directory


def enrich_metadata_endpoint(
    content_metadata: Dict[str, str], request: Optional[HttpRequest] = None
) -> Dict[str, str]:
    """Enrich content metadata dict with link to the upper metadata endpoint.

    Args:
        content_metadata: dict of data associated to a swh content metadata
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched content metadata dict filled with an additional url
    """
    c = content_metadata
    c["content_url"] = reverse(
        "api-1-content", url_args={"q": "sha1:%s" % c["id"]}, request=request
    )
    return c


def enrich_content(
    content: Dict[str, Any],
    request: Optional[HttpRequest] = None,
    top_url: Optional[bool] = False,
    query_string: Optional[str] = None,
) -> Dict[str, str]:
    """Enrich content with links to:
        - data_url: its raw data
        - filetype_url: its filetype information
        - language_url: its programming language information
        - license_url: its licensing information

    Args:
        content: dict of data associated to a swh content object
        top_url: whether or not to include the content url in
            the enriched data
        query_string: optional query string of type '<algo>:<hash>'
            used when requesting the content, it acts as a hint
            for picking the same hash method when computing
            the url listed above
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched content dict filled with additional urls

    """
    checksums = content
    if "checksums" in content:
        checksums = content["checksums"]
    hash_algo = "sha1"
    if query_string:
        hash_algo = parse_hash(query_string)[0]
    if hash_algo in checksums:
        q = "%s:%s" % (hash_algo, checksums[hash_algo])
        if top_url:
            content["content_url"] = reverse("api-1-content", url_args={"q": q})
        content["data_url"] = reverse(
            "api-1-content-raw", url_args={"q": q}, request=request
        )
        content["filetype_url"] = reverse(
            "api-1-content-filetype", url_args={"q": q}, request=request
        )
        content["language_url"] = reverse(
            "api-1-content-language", url_args={"q": q}, request=request
        )
        content["license_url"] = reverse(
            "api-1-content-license", url_args={"q": q}, request=request
        )

    return content


def enrich_revision(
    revision: Dict[str, Any], request: Optional[HttpRequest] = None
) -> Dict[str, Any]:
    """Enrich revision with links where it makes sense (directory, parents).
    Keep track of the navigation breadcrumbs if they are specified.

    Args:
        revision: the revision as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched revision dict filled with additional urls
    """

    revision["url"] = reverse(
        "api-1-revision", url_args={"sha1_git": revision["id"]}, request=request
    )
    revision["history_url"] = reverse(
        "api-1-revision-log", url_args={"sha1_git": revision["id"]}, request=request
    )

    if "directory" in revision:
        revision["directory_url"] = reverse(
            "api-1-directory",
            url_args={"sha1_git": revision["directory"]},
            request=request,
        )

    if "parents" in revision:
        parents = []
        for parent in revision["parents"]:
            parents.append(
                {
                    "id": parent,
                    "url": reverse(
                        "api-1-revision", url_args={"sha1_git": parent}, request=request
                    ),
                }
            )

        revision["parents"] = tuple(parents)

    if "children" in revision:
        children = []
        for child in revision["children"]:
            children.append(
                reverse("api-1-revision", url_args={"sha1_git": child}, request=request)
            )
        revision["children_urls"] = children

    if "decoding_failures" in revision and "message" in revision["decoding_failures"]:
        revision["message_url"] = reverse(
            "api-1-revision-raw-message",
            url_args={"sha1_git": revision["id"]},
            request=request,
        )

    return revision


def enrich_snapshot(
    snapshot: Dict[str, Any], request: Optional[HttpRequest] = None
) -> Dict[str, Any]:
    """Enrich snapshot with links to the branch targets

    Args:
        snapshot: the snapshot as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched snapshot dict filled with additional urls
    """
    if "branches" in snapshot:
        snapshot["branches"] = {
            k: enrich_object(v, request) if v else None
            for k, v in snapshot["branches"].items()
        }
        for k, v in snapshot["branches"].items():
            if v and v["target_type"] == "alias":
                branch = resolve_branch_alias(snapshot, v)
                if branch:
                    branch = enrich_object(branch, request)
                    v["target_url"] = branch["target_url"]
    return snapshot


def enrich_origin(
    origin: Union[Dict[str, Any], OriginInfo], request: Optional[HttpRequest] = None
) -> Dict[str, Any]:
    """Enrich origin dict with link to its visits

    Args:
        origin: the origin as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched origin dict filled with additional urls
    """
    origin_dict = dict(origin)
    if "url" in origin_dict:
        origin_dict["origin_visits_url"] = reverse(
            "api-1-origin-visits",
            url_args={"origin_url": origin_dict["url"]},
            request=request,
        )
        origin_dict["metadata_authorities_url"] = reverse(
            "api-1-raw-extrinsic-metadata-swhid-authorities",
            url_args={"target": Origin(url=origin_dict["url"]).swhid()},
            request=request,
        )

    return origin_dict


def enrich_origin_search_result(
    origin_search_result: Tuple[List[Dict[str, Any]], Optional[str], Optional[int]],
    request: Optional[HttpRequest] = None,
) -> Tuple[List[Dict[str, Any]], Optional[str], Optional[int]]:
    """Enrich origin search result with additional links

    Args:
        origin_search_result: tuple returned when searching origins
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched origin search result filled with additional urls
    """
    origins, page_token, total_results = origin_search_result
    return (
        [enrich_origin(origin, request=request) for origin in origins],
        page_token,
        total_results,
    )


def enrich_origin_visit(
    origin_visit: Dict[str, Any],
    request: Optional[HttpRequest] = None,
    with_origin_link: bool = False,
    with_origin_visit_link: bool = False,
) -> Dict[str, Any]:
    """Enrich origin visit dict with additional links

    Args:
        origin_visit: the origin visit as a dict
        with_origin_link: whether to add link to origin
        with_origin_visit_link: whether to add link to origin visit
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched origin visit dict filled with additional urls
    """
    ov = origin_visit
    if with_origin_link:
        ov["origin_url"] = reverse(
            "api-1-origin", url_args={"origin_url": ov["origin"]}, request=request
        )
    if with_origin_visit_link:
        ov["origin_visit_url"] = reverse(
            "api-1-origin-visit",
            url_args={"origin_url": ov["origin"], "visit_id": ov["visit"]},
            request=request,
        )
    snapshot = ov.get("snapshot")
    if snapshot:
        ov["snapshot_url"] = reverse(
            "api-1-snapshot", url_args={"snapshot_id": snapshot}, request=request
        )
    else:
        ov["snapshot"] = None
        ov["snapshot_url"] = None
    return ov


def enrich_extid(extid: Dict[str, Any], request: Optional[HttpRequest] = None):
    """Enrich extid dict with additional link

    Args:
        extid: the extid as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched extid dict filled with additional url
    """
    extid["target_url"] = reverse(
        "browse-swhid", url_args={"swhid": extid["target"]}, request=request
    )
    return extid



# Copyright (C) 2015-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from urllib.parse import unquote

import pytest

from django.conf import settings

from swh.web.save_code_now.models import (
    SAVE_REQUEST_ACCEPTED,
    SAVE_REQUEST_PENDING,
    SAVE_REQUEST_REJECTED,
    SAVE_TASK_PENDING,
    SaveAuthorizedOrigin,
    SaveOriginRequest,
    SaveUnauthorizedOrigin,
)
from swh.web.save_code_now.origin_save import can_save_origin
from swh.web.tests.helpers import check_http_get_response, check_http_post_response
from swh.web.utils import reverse

_authorized_origin_url = "https://scm.ourproject.org/anonscm/"
_unauthorized_origin_url = "https://www.softwareheritage.org/"


pytestmark = pytest.mark.django_db


@pytest.fixture(autouse=True)
def populated_db():
    SaveAuthorizedOrigin.objects.create(url=_authorized_origin_url)
    SaveUnauthorizedOrigin.objects.create(url=_unauthorized_origin_url)


def check_not_login(client, url):
    login_url = reverse(settings.LOGIN_URL, query_params={"next": url})

    resp = check_http_post_response(client, url, status_code=302)
    assert unquote(resp.url) == login_url


def test_add_authorized_origin_url(client, staff_user):
    authorized_url = "https://scm.adullact.net/anonscm/"
    assert can_save_origin(authorized_url) == SAVE_REQUEST_PENDING

    url = reverse(
        "admin-origin-save-add-authorized-url", url_args={"origin_url": authorized_url}
    )

    check_not_login(client, url)

    assert can_save_origin(authorized_url) == SAVE_REQUEST_PENDING

    client.force_login(staff_user)

    check_http_post_response(client, url, status_code=200)
    assert can_save_origin(authorized_url) == SAVE_REQUEST_ACCEPTED


def test_remove_authorized_origin_url(client, staff_user):
    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_ACCEPTED

    url = reverse(
        "admin-origin-save-remove-authorized-url",
        url_args={"origin_url": _authorized_origin_url},
    )

    check_not_login(client, url)

    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_ACCEPTED

    client.force_login(staff_user)
    check_http_post_response(client, url, status_code=200)
    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_PENDING


def test_add_unauthorized_origin_url(client, staff_user):
    unauthorized_url = "https://www.yahoo./"
    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_PENDING

    url = reverse(
        "admin-origin-save-add-unauthorized-url",
        url_args={"origin_url": unauthorized_url},
    )

    check_not_login(client, url)

    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_PENDING

    client.force_login(staff_user)
    check_http_post_response(client, url, status_code=200)
    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_REJECTED


def test_remove_unauthorized_origin_url(client, staff_user):
    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_REJECTED

    url = reverse(
        "admin-origin-save-remove-unauthorized-url",
        url_args={"origin_url": _unauthorized_origin_url},
    )

    check_not_login(client, url)

    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_REJECTED

    client.force_login(staff_user)
    check_http_post_response(client, url, status_code=200)
    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_PENDING


def test_accept_pending_save_request(client, staff_user, swh_scheduler):
    visit_type = "git"
    origin_url = "https://v2.pikacode.com/bthate/botlib.git"
    save_request_url = reverse(
        "api-1-save-origin",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )
    response = check_http_post_response(client, save_request_url, status_code=200)
    assert response.data["save_request_status"] == SAVE_REQUEST_PENDING

    accept_request_url = reverse(
        "admin-origin-save-request-accept",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    check_not_login(client, accept_request_url)

    client.force_login(staff_user)
    response = check_http_post_response(client, accept_request_url, status_code=200)

    response = check_http_get_response(client, save_request_url, status_code=200)
    assert response.data[0]["save_request_status"] == SAVE_REQUEST_ACCEPTED
    assert response.data[0]["save_task_status"] == SAVE_TASK_PENDING


def test_reject_pending_save_request(client, staff_user, swh_scheduler):
    visit_type = "git"
    origin_url = "https://wikipedia.com"

    save_request_url = reverse(
        "api-1-save-origin",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    response = check_http_post_response(client, save_request_url, status_code=200)
    assert response.data["save_request_status"] == SAVE_REQUEST_PENDING

    reject_request_url = reverse(
        "admin-origin-save-request-reject",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    check_not_login(client, reject_request_url)

    client.force_login(staff_user)
    response = check_http_post_response(client, reject_request_url, status_code=200)

    response = check_http_get_response(client, save_request_url, status_code=200)
    assert response.data[0]["save_request_status"] == SAVE_REQUEST_REJECTED
    assert response.data[0]["note"] == ""


def test_reject_pending_save_request_not_found(client, staff_user, swh_scheduler):
    visit_type = "git"
    origin_url = "https://example.org"

    reject_request_url = reverse(
        "admin-origin-save-request-reject",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    client.force_login(staff_user)
    check_http_post_response(client, reject_request_url, status_code=404)


def test_reject_pending_save_request_with_note(client, staff_user, swh_scheduler):
    visit_type = "git"
    origin_url = "https://wikipedia.com"

    save_request_url = reverse(
        "api-1-save-origin",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    response = check_http_post_response(client, save_request_url, status_code=200)
    assert response.data["save_request_status"] == SAVE_REQUEST_PENDING

    reject_request_url = reverse(
        "admin-origin-save-request-reject",
        url_args={"visit_type": visit_type, "origin_url": origin_url},
    )

    data = {"note": "The URL does not target a git repository"}

    client.force_login(staff_user)
    response = check_http_post_response(
        client, reject_request_url, status_code=200, data=data
    )

    response = check_http_get_response(client, save_request_url, status_code=200)
    assert response.data[0]["save_request_status"] == SAVE_REQUEST_REJECTED
    assert response.data[0]["note"] == data["note"]


def test_remove_save_request(client, staff_user):
    sor = SaveOriginRequest.objects.create(
        visit_type="git",
        origin_url="https://wikipedia.com",
        status=SAVE_REQUEST_PENDING,
    )
    assert SaveOriginRequest.objects.count() == 1

    remove_request_url = reverse(
        "admin-origin-save-request-remove", url_args={"sor_id": sor.id}
    )

    check_not_login(client, remove_request_url)

    client.force_login(staff_user)
    check_http_post_response(client, remove_request_url, status_code=200)
    assert SaveOriginRequest.objects.count() == 0



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.SeparateDatabaseAndState(
            # as we move the mailmap feature to a dedicated django application,
            # no need to recreate database tables as they already exist
            state_operations=[
                migrations.CreateModel(
                    name="UserMailmap",
                    fields=[
                        (
                            "id",
                            models.AutoField(
                                auto_created=True,
                                primary_key=True,
                                serialize=False,
                                verbose_name="ID",
                            ),
                        ),
                        ("user_id", models.CharField(max_length=50, null=True)),
                        ("from_email", models.TextField(unique=True)),
                        ("from_email_verified", models.BooleanField(default=False)),
                        (
                            "from_email_verification_request_date",
                            models.DateTimeField(null=True),
                        ),
                        ("display_name", models.TextField()),
                        ("display_name_activated", models.BooleanField(default=False)),
                        ("to_email", models.TextField(null=True)),
                        ("to_email_verified", models.BooleanField(default=False)),
                        (
                            "to_email_verification_request_date",
                            models.DateTimeField(null=True),
                        ),
                        (
                            "mailmap_last_processing_date",
                            models.DateTimeField(null=True),
                        ),
                        ("last_update_date", models.DateTimeField(auto_now=True)),
                    ],
                    options={
                        "db_table": "user_mailmap",
                    },
                ),
                migrations.CreateModel(
                    name="UserMailmapEvent",
                    fields=[
                        (
                            "id",
                            models.AutoField(
                                auto_created=True,
                                primary_key=True,
                                serialize=False,
                                verbose_name="ID",
                            ),
                        ),
                        ("timestamp", models.DateTimeField(auto_now=True)),
                        ("user_id", models.CharField(max_length=50)),
                        ("request_type", models.CharField(max_length=50)),
                        ("request", models.TextField()),
                        ("successful", models.BooleanField(default=False)),
                    ],
                    options={
                        "db_table": "user_mailmap_event",
                    },
                ),
                migrations.AddIndex(
                    model_name="usermailmapevent",
                    index=models.Index(
                        fields=["timestamp"], name="user_mailma_timesta_1f7aef_idx"
                    ),
                ),
            ],
            database_operations=[],
        ),
    ]



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from base64 import b64encode

import pytest

from django.conf import settings

from swh.web.auth.utils import (
    ADMIN_LIST_DEPOSIT_PERMISSION,
    get_or_create_django_permission,
)
from swh.web.config import get_config
from swh.web.tests.helpers import check_html_get_response, check_http_get_response
from swh.web.utils import reverse


def test_deposit_admin_view_not_available_for_anonymous_user(client):
    url = reverse("admin-deposit")
    resp = check_html_get_response(client, url, status_code=302)
    assert resp["location"] == reverse(settings.LOGIN_URL, query_params={"next": url})


@pytest.mark.django_db
def test_deposit_admin_view_available_for_staff_user(client, staff_user):
    client.force_login(staff_user)
    url = reverse("admin-deposit")
    check_html_get_response(
        client, url, status_code=200, template_used="deposit-admin.html"
    )


@pytest.mark.django_db
def test_deposit_admin_view_available_for_user_with_permission(client, regular_user):
    regular_user.user_permissions.add(
        get_or_create_django_permission(ADMIN_LIST_DEPOSIT_PERMISSION)
    )
    client.force_login(regular_user)
    url = reverse("admin-deposit")
    check_html_get_response(
        client, url, status_code=200, template_used="deposit-admin.html"
    )


@pytest.mark.django_db
def test_deposit_admin_view_list_deposits(
    client, staff_user, requests_mock, config_updater
):
    deposits_data = {
        "data": [
            {
                "external_id": "hal-02527986",
                "id": 1066,
                "raw_metadata": None,
                "reception_date": "2022-04-08T14:12:34.143000Z",
                "status": "rejected",
                "status_detail": None,
                "swhid": None,
                "swhid_context": None,
                "type": "code",
                "uri": "https://inria.halpreprod.archives-ouvertes.fr/hal-02527986",
            },
            {
                "external_id": "hal-01243573",
                "id": 1065,
                "raw_metadata": None,
                "reception_date": "2022-04-08T12:53:50.940000Z",
                "status": "rejected",
                "status_detail": None,
                "swhid": None,
                "swhid_context": None,
                "type": "code",
                "uri": "https://inria.halpreprod.archives-ouvertes.fr/hal-01243573",
            },
        ],
        "draw": 2,
        "recordsFiltered": 645,
        "recordsTotal": 1066,
    }

    config_updater(
        {
            "deposit": {
                "private_api_url": "http://deposit.example.org/1/private",
                "private_api_user": "swhworker",
                "private_api_password": "some-password",
            }
        }
    )

    config = get_config()["deposit"]
    private_api_url = config["private_api_url"].rstrip("/") + "/"
    deposits_list_url = private_api_url + "deposits/datatables/"

    basic_auth_payload = (
        config["private_api_user"] + ":" + config["private_api_password"]
    ).encode()

    requests_mock.get(
        deposits_list_url,
        json=deposits_data,
        request_headers={
            "Authorization": f"Basic {b64encode(basic_auth_payload).decode('ascii')}"
        },
    )

    client.force_login(staff_user)
    url = reverse("admin-deposit-list")
    check_http_get_response(
        client, url, status_code=200, content_type="application/json"
    )



# Copyright (C) 2015-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import io
import os
from typing import Optional

from django.http import FileResponse
from rest_framework import serializers
from rest_framework.request import Request

from swh.web.api import utils
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup
from swh.web.utils import archive
from swh.web.utils.exc import NotFoundExc


class ContentRawQuerySerializer(serializers.Serializer):
    """Content Raw query parameters serializer."""

    filename = serializers.CharField(
        required=False, min_length=1, max_length=255, default=""
    )


@api_route(
    r"/content/(?P<q>[0-9a-z_:]*[0-9a-f]+)/filetype/",
    "api-1-content-filetype",
    checksum_args=["q"],
)
@api_doc("/content/filetype/", category="Metadata")
@format_docstring()
def api_content_filetype(request: Request, q: str):
    """
    .. http:get:: /api/1/content/[(hash_type):](hash)/filetype/

        Get information about the detected MIME type of a content object.

        :param string hash_type: optional parameter specifying which hashing algorithm
            has been used to compute the content checksum. It can be either ``sha1``,
            ``sha1_git``, ``sha256`` or ``blake2s256``. If that parameter is not
            provided, it is assumed that the hashing algorithm used is ``sha1``.
        :param string hash: hexadecimal representation of the checksum value computed
            with the specified hashing algorithm.

        :>json object content_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/` for getting information
            about the content
        :>json string encoding: the detected content encoding
        :>json string id: the **sha1** identifier of the content
        :>json string mimetype: the detected MIME type of the content
        :>json object tool: information about the tool used to detect the content
            filetype

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested content cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/sha1:dc2830a9e72f23c1dfebef4413003221baa5fb62/filetype/`
    """
    return api_lookup(
        archive.lookup_content_filetype,
        q,
        notfound_msg="No filetype information found for content {}.".format(q),
        enrich_fn=utils.enrich_metadata_endpoint,
        request=request,
    )


@api_route(
    r"/content/(?P<q>[0-9a-z_:]*[0-9a-f]+)/language/",
    "api-1-content-language",
    checksum_args=["q"],
)
@api_doc("/content/language/", category="Metadata")
@format_docstring()
def api_content_language(request: Request, q: str):
    """
    .. http:get:: /api/1/content/[(hash_type):](hash)/language/

        Get information about the programming language used in a content object.

        Note: this endpoint currently returns no data.

        :param string hash_type: optional parameter specifying which hashing algorithm
            has been used to compute the content checksum. It can be either ``sha1``,
            ``sha1_git``, ``sha256`` or ``blake2s256``. If that parameter is not
            provided, it is assumed that the hashing algorithm used is ``sha1``.
        :param string hash: hexadecimal representation of the checksum value computed
            with the specified hashing algorithm.

        :>json object content_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/` for getting information
            about the content
        :>json string id: the **sha1** identifier of the content
        :>json string lang: the detected programming language if any
        :>json object tool: information about the tool used to detect the programming
            language

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested content cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/sha1:dc2830a9e72f23c1dfebef4413003221baa5fb62/language/`
    """
    return api_lookup(
        archive.lookup_content_language,
        q,
        notfound_msg="No language information found for content {}.".format(q),
        enrich_fn=utils.enrich_metadata_endpoint,
        request=request,
    )


@api_route(
    r"/content/(?P<q>[0-9a-z_:]*[0-9a-f]+)/license/",
    "api-1-content-license",
    checksum_args=["q"],
)
@api_doc("/content/license/", category="Metadata")
@format_docstring()
def api_content_license(request: Request, q: str):
    """
    .. http:get:: /api/1/content/[(hash_type):](hash)/license/

        Get information about the license of a content object.

        :param string hash_type: optional parameter specifying which hashing algorithm
            has been used to compute the content checksum. It can be either ``sha1``,
            ``sha1_git``, ``sha256`` or ``blake2s256``. If that parameter is not
            provided, it is assumed that the hashing algorithm used is ``sha1``.
        :param string hash: hexadecimal representation of the checksum value computed
            with the specified hashing algorithm.

        :>json object content_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/` for getting information
            about the content
        :>json string id: the **sha1** identifier of the content
        :>json array licenses: array of strings containing the detected license names
        :>json object tool: information about the tool used to detect the license

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested content cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/sha1:dc2830a9e72f23c1dfebef4413003221baa5fb62/license/`
    """
    return api_lookup(
        archive.lookup_content_license,
        q,
        notfound_msg="No license information found for content {}.".format(q),
        enrich_fn=utils.enrich_metadata_endpoint,
        request=request,
    )


@api_route(
    r"/content/(?P<q>[0-9a-z_:]*[0-9a-f]+)/raw/",
    "api-1-content-raw",
    checksum_args=["q"],
    query_params_serializer=ContentRawQuerySerializer,
)
@api_doc("/content/raw/", category="Archive")
def api_content_raw(request: Request, q: str, validated_query_params: dict[str, str]):
    """
    .. http:get:: /api/1/content/[(hash_type):](hash)/raw/

        Get the raw content of a content object (aka a "blob"), as a byte sequence.

        :param string hash_type: optional parameter specifying which hashing algorithm
            has been used to compute the content checksum. It can be either ``sha1``,
            ``sha1_git``, ``sha256`` or ``blake2s256``. If that parameter is not
            provided, it is assumed that the hashing algorithm used is ``sha1``.
        :param string hash: hexadecimal representation of the checksum value computed
            with the specified hashing algorithm.
        :query string filename: if provided, the downloaded content will get that
            filename

        :resheader Content-Type: application/octet-stream

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested content cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/sha1:dc2830a9e72f23c1dfebef4413003221baa5fb62/raw/`
    """
    content = archive.lookup_content(q, with_data=True)
    if not content:
        raise NotFoundExc("Content %s is not found." % q)

    filename = validated_query_params["filename"] or "content_%s_raw" % q.replace(
        ":", "_"
    )

    return FileResponse(
        io.BytesIO(content["data"]),  # not copied, as this is never modified
        filename=os.path.basename(filename),
        content_type="application/octet-stream",
        as_attachment=True,
    )


@api_route(r"/content/known/search/", "api-1-content-known", methods=["POST"])
@api_route(r"/content/known/(?P<q>(?!search).+)/", "api-1-content-known")
@api_doc("/content/known/", category="Archive", tags=["hidden"])
@format_docstring()
def api_check_content_known(request: Request, q: Optional[str] = None):
    """
    .. http:get:: /api/1/content/known/(sha1)[,(sha1), ...,(sha1)]/

        Check whether some content(s) (aka "blob(s)") is present in the archive
        based on its **sha1** checksum.

        :param string sha1: hexadecimal representation of the **sha1** checksum value
            for the content to check existence. Multiple values can be provided
            separated by ','.

        {common_headers}

        :>json array search_res: array holding the search result for each provided
            **sha1**
        :>json object search_stats: some statistics regarding the number of **sha1**
            provided and the percentage of those found in the archive

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1** has been provided

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/known/dc2830a9e72f23c1dfebef4413003221baa5fb62,0c3f19cb47ebfbe643fb19fa94c874d18fa62d12/`
    """
    search_stats = {"nbfiles": 0, "pct": 0}
    search_res = None

    queries = []
    # GET: Many hash separated values request
    if q:
        hashes = q.split(",")
        for v in hashes:
            queries.append({"filename": None, "sha1": v})

    # POST: Many hash requests in post form submission
    elif request.method == "POST":
        data = request.data
        # Remove potential inputs with no associated value
        for k, v in data.items():
            if v is not None:
                if k == "q" and len(v) > 0:
                    queries.append({"filename": None, "sha1": v})
                elif v != "":
                    queries.append({"filename": k, "sha1": v})

    if queries:
        lookup = archive.lookup_multiple_hashes(queries)
        result = []
        nb_queries = len(queries)
        for el in lookup:
            res_d = {"sha1": el["sha1"], "found": el["found"]}
            if "filename" in el and el["filename"]:
                res_d["filename"] = el["filename"]
            result.append(res_d)
            search_res = result
            nbfound = len([x for x in lookup if x["found"]])
            search_stats["nbfiles"] = nb_queries
            search_stats["pct"] = int((nbfound / nb_queries) * 100)

    return {"search_res": search_res, "search_stats": search_stats}


@api_route(
    r"/content/(?P<q>[0-9a-z_:]*[0-9a-f]+)/", "api-1-content", checksum_args=["q"]
)
@api_doc("/content/", category="Archive")
@format_docstring()
def api_content_metadata(request: Request, q: str):
    """
    .. http:get:: /api/1/content/[(hash_type):](hash)/

        Get information about a content (aka a "blob") object.
        In the archive, a content object is identified based on checksum
        values computed using various hashing algorithms.

        :param string hash_type: optional parameter specifying which hashing algorithm
            has been used to compute the content checksum. It can be either ``sha1``,
            ``sha1_git``, ``sha256`` or ``blake2s256``. If that parameter is not
            provided, it is assumed that the hashing algorithm used is ``sha1``.
        :param string hash: hexadecimal representation of the checksum value computed
            with the specified hashing algorithm.

        {common_headers}

        :>json object checksums: object holding the computed checksum values for the
            requested content
        :>json string data_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/raw/`
            for downloading the content raw bytes
        :>json string filetype_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/filetype/`
            for getting information about the content MIME type
        :>json string language_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/language/`
            for getting information about the programming language used in the content
        :>json number length: length of the content in bytes
        :>json string license_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/license/`
            for getting information about the license of the content

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested content cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/`
    """
    return api_lookup(
        archive.lookup_content,
        q,
        notfound_msg="Content with {} not found.".format(q),
        enrich_fn=functools.partial(utils.enrich_content, query_string=q),
        request=request,
    )



# Copyright (C) 2015-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from types import GeneratorType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from django.http import HttpRequest
from rest_framework.decorators import api_view
from rest_framework.request import Request
from rest_framework.response import Response

from swh.web.api.apiurls import api_route, api_urls
from swh.web.utils.exc import NotFoundExc

EnrichFunction = Callable[[Dict[str, str], Optional[HttpRequest]], Dict[str, str]]

EnrichFunctionSearchResult = Callable[
    [Tuple[List[Dict[str, Any]], Optional[str], Optional[int]], Optional[HttpRequest]],
    Tuple[List[Dict[str, Any]], Optional[str], Optional[int]],
]


def api_lookup(
    lookup_fn: Callable[..., Any],
    *args: Any,
    notfound_msg: Optional[str] = "Object not found",
    enrich_fn: Optional[Union[EnrichFunction, EnrichFunctionSearchResult]] = None,
    request: Optional[HttpRequest] = None,
    **kwargs: Any,
):
    r"""
    Capture a redundant behavior of:
        - looking up the backend with a criteria (be it an identifier or
          checksum) passed to the function lookup_fn
        - if nothing is found, raise :class:`swh.web.utils.exc.NotFoundExc`
          exception with error message ``notfound_msg``.
        - Otherwise if something is returned:
            - either as list, map or generator, map the enrich_fn function to
              it and return the resulting data structure as list.
            - either as dict and pass to enrich_fn and return the dict
              enriched.

    Args:
        - lookup_fn: function expects one criteria and optional supplementary
          \*args.
        - \*args: supplementary arguments to pass to lookup_fn.
        - notfound_msg: if nothing matching the criteria is found,
          raise :class:`swh.web.utils.exc.NotFoundExc` with this error message.
        - enrich_fn: Function to use to enrich the result returned by
          lookup_fn. Default to the identity function if not provided.
        - request: Input HTTP request that will be provided as parameter
          to enrich_fn.


    Raises:
        swh.web.utils.exc.NotFoundExc: when a requested object is not found
            or whatever ``lookup_fn`` raises.

    """

    def _enrich_fn_noop(x, request):
        return x

    if enrich_fn is None:
        enrich_fn = _enrich_fn_noop
    res = lookup_fn(*args, **kwargs)
    if res is None:
        raise NotFoundExc(notfound_msg)
    if isinstance(res, (list, GeneratorType)) or type(res) is map:
        return [enrich_fn(x, request) for x in iter(res)]
    return enrich_fn(res, request)


@api_view(["GET", "HEAD"])
def api_home(request: Request):
    return Response({}, template_name="api.html")


api_urls.add_url_pattern("api/", api_home, view_name="api-1-homepage")


@api_route(r"/", "api-1-endpoints")
def api_endpoints(request):
    """Display the list of opened api endpoints."""
    routes_by_category = {}
    for route, doc in api_urls.get_app_endpoints().items():
        doc["doc_intro"] = doc["docstring"].split("\n\n")[0]
        routes_by_category.setdefault(doc["category"], []).append(doc)

    for routes in routes_by_category.values():
        routes.sort(key=lambda route: route["route"])

    # sort routes by alphabetical category name, with 'miscellaneous' at the end
    misc_routes = routes_by_category.pop("Miscellaneous")
    sorted_routes = sorted(routes_by_category.items())
    sorted_routes.append(("Miscellaneous", misc_routes))

    env = {"doc_routes": sorted_routes}
    return Response(env, template_name="api-endpoints.html")
